from .utils import (
    check_below_height_condition,
    check_above_height_condition,
    check_above_light_level_condition,
    check_below_light_level_condition,
)

# TODO (yunfan): complete these dicts
# this dict specifies target items and corresponding blocks to obtain,
# e.g., break diamond ore to obtain a diamond
Target2SpawnItem = {
    # ore
    "cobblestone": "stone",
    "diamond": "diamond_ore",
    "redstone": "redstone_ore",
    "coal": "coal_ore",
    # ore byproduct
    "iron_ingot": "iron_ore",
    "gold_ingot": "gold_ore",
    # mob byproduct
    "leather": "cow",
    "mutton": "sheep",
    "porkchop": "pig",
    "beef": "cow",
    "bone": "skeleton",
    "web": "spider",
    "feather": "chicken",
    "wool": "sheep",
    "milk_bucket": "cow",
    # plant byproduct
    "wheat_seeds": "wheat",
    "beetroot_seeds": "beetroot",
    "pumpkin_seeds": "pumpkin",
}

# this dict specifies spawn conditions
SpawnItem2Condition = {
    # ore
    "stone": check_below_height_condition(height_threshold=60),
    "coal_ore": check_below_height_condition(height_threshold=60),
    "iron_ore": check_below_height_condition(height_threshold=50),
    "gold_ore": check_below_height_condition(height_threshold=29),
    "diamond_ore": check_below_height_condition(height_threshold=14),
    "redstone_ore": check_below_height_condition(height_threshold=16),
    # natural items
    "pumpkin": check_above_height_condition(height_threshold=62),
    "reeds": check_above_height_condition(height_threshold=62),
    "wheat": check_above_height_condition(height_threshold=62),
    "beetroot": check_above_height_condition(height_threshold=62),
    "potato": check_above_height_condition(height_threshold=62),
    "carrot": check_above_height_condition(height_threshold=62),
    # night mobs
    "zombie": check_below_light_level_condition(light_level_threshold=7),
    "creeper": check_below_light_level_condition(light_level_threshold=7),
    "spider": check_below_light_level_condition(light_level_threshold=7),
    "skeleton": check_below_light_level_condition(light_level_threshold=7),
    "witch": check_below_light_level_condition(light_level_threshold=7),
    "enderman": check_below_light_level_condition(light_level_threshold=7),
    "bat": check_below_light_level_condition(light_level_threshold=7),
    "husk": check_below_light_level_condition(light_level_threshold=7),
    "slime": check_below_light_level_condition(light_level_threshold=7),
    # day mobs
    "pig": check_above_light_level_condition(light_level_threshold=9),
    "cow": check_above_light_level_condition(light_level_threshold=9),
    "chicken": check_above_light_level_condition(light_level_threshold=9),
    "sheep": check_above_light_level_condition(light_level_threshold=9),
    "rabbit": check_above_light_level_condition(light_level_threshold=9),
    "horse": check_above_light_level_condition(light_level_threshold=9),
    "donkey": check_above_light_level_condition(light_level_threshold=9),
    "mooshroom": check_above_light_level_condition(light_level_threshold=9),
    "llama": check_above_light_level_condition(light_level_threshold=7),
}



# Copyright (c) 2020 All Rights Reserved
# Author: William H. Guss, Brandon Houghton
import numpy as np

import minedojo.sim.mc_meta.mc as mc
import minedojo.sim.spaces as spaces
from minedojo.sim.handlers.translation import (
    KeymapTranslationHandler,
    TranslationHandlerGroup,
)


__all__ = ["ObservationFromLifeStats"]


class ObservationFromLifeStats(TranslationHandlerGroup):
    """Groups all lifestats observations together to correspond to one XML element."""

    def to_string(self) -> str:
        return "life_stats"

    def __init__(self):
        super(ObservationFromLifeStats, self).__init__(
            handlers=[
                _LifeObservation(),
                _ArmorObservation(),
                _FoodObservation(),
                _SaturationObservation(),
                _XPObservation(),
                _BreathObservation(),
                _IsSleepingObservation(),
            ]
        )

    def xml_template(self) -> str:
        return str("""<ObservationFromFullStats/>""")


class LifeStatsObservation(KeymapTranslationHandler):
    def to_hero(self, x) -> str:
        pass

    def __init__(self, hero_keys, univ_keys, space, default_if_missing=None):
        self.hero_keys = hero_keys
        self.univ_keys = univ_keys
        super().__init__(
            hero_keys=hero_keys,
            univ_keys=["life_stats"] + univ_keys,
            space=space,
            default_if_missing=default_if_missing,
        )

    def xml_template(self) -> str:
        return str("""<ObservationFromFullStats/>""")


class _LifeObservation(LifeStatsObservation):
    """
    Handles life observation / health observation. Its initial value on world creation is 20 (full bar)
    """

    def __init__(self):
        keys = ["life"]
        super().__init__(
            hero_keys=keys,
            univ_keys=keys,
            space=spaces.Box(low=0, high=mc.MAX_LIFE, shape=(1,), dtype=np.float32),
            default_if_missing=mc.MAX_LIFE,
        )

    def from_hero(self, hero_dict, dtype=None):
        return np.expand_dims(super().from_hero(hero_dict, dtype), axis=0).astype(
            self.space.dtype
        )


class _ArmorObservation(LifeStatsObservation):
    """
    Handles life observation / armor observation. Its initial value on world creation is 0 (wearing no armors)
    """

    def __init__(self):
        keys = ["armor"]
        super().__init__(
            hero_keys=keys,
            univ_keys=keys,
            space=spaces.Box(low=0, high=mc.MAX_ARMOR, shape=(1,), dtype=np.float32),
            default_if_missing=0,
        )

    def from_hero(self, hero_dict, dtype=None):
        return np.expand_dims(super().from_hero(hero_dict, dtype), axis=0).astype(
            self.space.dtype
        )


class _FoodObservation(LifeStatsObservation):
    """
    Handles food_level observation representing the player's current hunger level, shown on the hunger bar. Its initial
    value on world creation is 20 (full bar) - https://minecraft.gamepedia.com/Hunger#Mechanics
    """

    def __init__(self):
        super().__init__(
            hero_keys=["food"],
            univ_keys=["food"],
            space=spaces.Box(low=0, high=mc.MAX_FOOD, shape=(1,), dtype=np.float32),
            default_if_missing=mc.MAX_FOOD,
        )

    def from_hero(self, hero_dict, dtype=None):
        return np.expand_dims(super().from_hero(hero_dict, dtype), axis=0).astype(
            self.space.dtype
        )


class _SaturationObservation(LifeStatsObservation):
    """
    Returns the food saturation observation which determines how fast the hunger level depletes and is controlled by the
    kinds of food the player has eaten. Its maximum value always equals foodLevel's value and decreases with the hunger
    level. Its initial value on world creation is 5. - https://minecraft.gamepedia.com/Hunger#Mechanics
    """

    def __init__(self):
        super().__init__(
            hero_keys=["saturation"],
            univ_keys=["saturation"],
            space=spaces.Box(
                low=0, high=mc.MAX_FOOD_SATURATION, shape=(1,), dtype=np.float32
            ),
            default_if_missing=5.0,
        )

    def from_hero(self, hero_dict, dtype=None):
        return np.expand_dims(super().from_hero(hero_dict, dtype), axis=0).astype(
            self.space.dtype
        )


class _XPObservation(LifeStatsObservation):
    """
    Handles observation of experience points 1395 experience corresponds with level 30
    - see https://minecraft.gamepedia.com/Experience for more details
    """

    def __init__(self):
        keys = ["xp"]
        super().__init__(
            hero_keys=keys,
            univ_keys=keys,
            space=spaces.Box(low=0, high=mc.MAX_XP, shape=(1,), dtype=np.float32),
            default_if_missing=0,
        )

    def from_hero(self, hero_dict, dtype=None):
        return np.expand_dims(super().from_hero(hero_dict, dtype), axis=0).astype(
            self.space.dtype
        )


class _BreathObservation(LifeStatsObservation):
    """
    Handles observation of breath which tracks the amount of air remaining before beginning to suffocate
    """

    def __init__(self):
        super().__init__(
            hero_keys=["air"],
            univ_keys=["air"],
            space=spaces.Box(low=0, high=mc.MAX_BREATH, shape=(1,), dtype=np.float32),
            default_if_missing=300,
        )

    def to_string(self) -> str:
        return "oxygen"

    def from_hero(self, hero_dict, dtype=None):
        return np.expand_dims(super().from_hero(hero_dict, dtype), axis=0).astype(
            self.space.dtype
        )


class _IsSleepingObservation(LifeStatsObservation):
    def __init__(self):
        super().__init__(
            hero_keys=["is_sleeping"],
            univ_keys=["is_sleeping"],
            space=spaces.Discrete(2),
            default_if_missing=0,
        )


class _IsAliveObservation(LifeStatsObservation):
    def __init__(self):
        keys = ["is_alive"]
        super().__init__(
            hero_keys=keys,
            univ_keys=keys,
            space=spaces.Discrete(2),
            default_if_missing=1,
        )



import os
import sys
import uuid
import time
import struct
import psutil
import socket
import atexit
import tempfile
import threading
import functools
import subprocess
import collections

import Pyro4
import locale
import shutil
import logging

from ...bridge import utils as U

__all__ = ["MinecraftInstance"]


logger = logging.getLogger(__name__)
MALMO_VERSION = "0.37.0"
MC_DIR = os.path.join(
    os.path.dirname(os.path.dirname(os.path.dirname(__file__))),
    "Malmo",
    "Minecraft",
)
SCHEMAS_DIR = os.path.join(
    os.path.dirname(os.path.dirname(os.path.dirname(__file__))),
    "Malmo",
    "Schemas",
)


@Pyro4.expose
class MinecraftInstance:
    """
    A subprocess wrapper which maintains a reference to a minecraft subprocess
    and also allows for stable closing and launching of such subprocesses
    across different platforms.

    The Minecraft instance class works by launching two subprocesses:
    the Malmo subprocess, and a watcher subprocess with access to
    the process IDs of both the parent process and the Malmo subprocess.
    If the parent process dies, it will kill the subprocess, and then itself.

    This scheme has a single failure point of the process dying before the watcher process is launched.
    """

    MAX_PIPE_LENGTH = 500

    def __init__(self, port=None, existing=False, seed=None, instance_id=None):
        """
        Launches the subprocess.
        """
        self.running = False
        self._starting = True
        self.minecraft_process = None
        self.watcher_process = None
        self.xml = None
        self.role = None
        self._client_socket = None
        self._port = port
        self._jdwp = None
        self._host = "localhost"
        self.locked = False
        self.uuid = str(uuid.uuid4()).replace("-", "")[:6]
        self.existing = existing
        self.minecraft_dir = None
        self.instance_dir = None
        self.owner = None
        self._had_to_clean = False
        self.instance_id = instance_id
        self._seed = seed
        self._target_port = port

        self._setup_logging()

    # class-wide Pyro4 exposes are not encouraged as of new versions
    @Pyro4.expose
    @property
    def had_to_clean(self):
        return self._had_to_clean

    @Pyro4.expose
    @had_to_clean.setter
    def had_to_clean(self, value):
        self._had_to_clean = value

    @Pyro4.expose
    @property
    def client_socket(self):
        return self._client_socket

    @Pyro4.expose
    @client_socket.setter
    def client_socket(self, value):
        # Not ideal as you should not be trying to serialize
        # sockets. However the code does set the socket to
        # Nones at times, so lets allow it.
        self._client_socket = value

    def create_instance_socket(self, socktime):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        sock.settimeout(socktime)
        sock.connect((self.host, self.port))
        self.client_socket = sock

    def client_socket_send_message(self, msg):
        U.send_message(self.client_socket, msg)

    def client_socket_recv_message(self):
        return U.recv_message(self.client_socket)

    def client_socket_close(self):
        self.client_socket.close()
        self.client_socket = None

    def client_socket_shutdown(self, param):
        self.client_socket.shutdown(param)
        self.client_socket = None

    def has_client_socket(self):
        return self.client_socket is not None

    @property
    def actor_name(self):
        return f"actor{self.role}"

    def launch(self, daemonize=False, replaceable=True):
        from ..utils import watchdog
        from .manager import InstanceManager

        port = self._target_port
        self._starting = True

        if not self.existing:
            if not port:
                port = InstanceManager._get_valid_port()

            self.instance_dir = tempfile.mkdtemp()
            self.minecraft_dir = os.path.join(self.instance_dir, "Minecraft")
            shutil.copytree(
                os.path.join(MC_DIR),
                self.minecraft_dir,
                ignore=shutil.ignore_patterns("**.lock"),
            )
            shutil.copytree(
                os.path.join(SCHEMAS_DIR),
                os.path.join(self.instance_dir, "Schemas"),
            )

            # 0. Get PID of launcher.
            parent_pid = os.getpid()
            # 1. Launch minecraft process
            self.minecraft_process = self._launch_minecraft(
                port,
                self.minecraft_dir,
                replaceable=replaceable,
            )

            # 2. Create a watcher process to ensure things get cleaned up
            if not daemonize:
                # 2. Create a watcher process to ensure things get cleaned up
                self.watcher_process = watchdog.launch(
                    parent_pid, self.minecraft_process.pid, self.instance_dir
                )

            # wait until Minecraft process has output "CLIENT enter state: DORMANT"
            lines = []
            client_ready = False
            server_ready = False

            while True:
                mine_log_encoding = locale.getpreferredencoding(False)
                line = self.minecraft_process.stdout.readline().decode(
                    mine_log_encoding
                )

                if os.environ.get("MINEDOJO_DEBUG_LOG", False):
                    # Print Java logs to console
                    print(line.strip("\n"))

                # Check for failures and print useful messages!
                # _check_for_launch_errors(line)

                if not line:
                    # IF THERE WAS AN ERROR STARTING THE MC PROCESS
                    # Print the whole logs!
                    error_str = ""
                    for l in lines:
                        spline = "\n".join(l.split("\n")[:-1])
                        self._logger.error(spline)
                        error_str += spline + "\n"
                    # Throw an exception!
                    raise EOFError(
                        error_str
                        + "\n\nMinecraft process finished unexpectedly. There was an error with Malmo."
                    )

                lines.append(line)
                self._log_heuristic("\n".join(line.split("\n")[:-1]))

                MALMOENVPORTSTR = "***** Start MalmoEnvServer on port "
                port_received = MALMOENVPORTSTR in line
                if port_received:
                    self._port = int(line.split(MALMOENVPORTSTR)[-1].strip())

                client_ready = "CLIENT enter state: DORMANT" in line
                server_ready = "SERVER enter state: DORMANT" in line

                if client_ready:
                    break

            if not self.port:
                raise RuntimeError(
                    "Malmo failed to start the MalmoEnv server! Check the logs from the Minecraft process."
                )
            self._logger.info("Minecraft process ready")

            if port != self._port:
                self._logger.warning(
                    f"Tried to launch Minecraft on port {port} but that port was taken, instead Minecraft is using port {self.port}."
                )

            # suppress entire output, otherwise the subprocess will block
            # NB! there will be still logs under Malmo/Minecraft/run/logs
            # FNULL = open(os.devnull, 'w')
            # launch a logger process
            def log_to_file(logdir):
                if not os.path.exists(os.path.join(logdir, "logs")):
                    os.makedirs((os.path.join(logdir, "logs")))

                file_path = os.path.join(
                    logdir, "logs", f"mc_{self._target_port - 9000}.log"
                )

                logger.info(f"Logging output of Minecraft to {file_path}")

                mine_log = open(file_path, "wb+")
                mine_log.truncate(0)
                mine_log_encoding = locale.getpreferredencoding(False)

                try:
                    while self.running:
                        line = self.minecraft_process.stdout.readline()
                        if not line:
                            break

                        try:
                            linestr = line.decode(mine_log_encoding)
                        except UnicodeDecodeError:
                            mine_log_encoding = locale.getpreferredencoding(False)
                            logger.error(
                                "UnicodeDecodeError, switching to default encoding"
                            )
                            linestr = line.decode(mine_log_encoding)

                        if os.environ.get("MINEDOJO_DEBUG_LOG", False):
                            # Print Java logs to console
                            print(linestr.strip("\n"))

                        linestr = "\n".join(linestr.split("\n")[:-1])
                        # some heuristics to figure out which messages
                        # need to be elevated in logging level
                        # "   at " elevation is related to logging exception's stacktrace
                        self._log_heuristic(linestr)
                        mine_log.write(line)
                        mine_log.flush()
                finally:
                    mine_log.close()

            logdir = os.environ.get("MALMO_MINECRAFT_OUTPUT_LOGDIR", ".")
            self.running = True
            self._logger_thread = threading.Thread(
                target=functools.partial(log_to_file, logdir=logdir)
            )
            self._logger_thread.setDaemon(True)
            self._logger_thread.start()
            self._starting = False

        else:
            assert port is not None, "No existing port specified."
            self._port = port
            self._starting = False
            self.running = True

        # Make a hook to kill
        if not daemonize:
            atexit.register(lambda: self._destruct())

    def kill(self):
        """
        Kills the process (if it has been launched.)
        """
        self._destruct()
        pass

    def close(self):
        """Closes the object."""
        self._destruct(should_close=True)

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port

    @property
    def jdwp_port(self):
        """
        JDWP (Java Debug Wire Protocol) port, if any, so the instance can be debugged with an
        attached debugger.
        """
        return self._jdwp

    @jdwp_port.setter
    def jdwp_port(self, port):
        """
        JDWP (Java Debug Wire Protocol) port, if any, so the instance can be debugged with an
        attached debugger.
        :param port: Port to set (0 or None to disable remote debugging).
        """
        self._jdwp = port

    def get_output(self):
        while self.running or self._starting:
            try:
                level, line = self._output_stream.pop()
                return (
                    (line.levelno, line.getMessage(), line.name),
                    self.running or self._starting,
                )
            except IndexError:
                time.sleep(0.5)
        else:
            return None, False

    def _setup_logging(self):
        # Set up an output stream handler.
        self._logger = logging.getLogger(f"{__name__}.instance.{self.uuid}")
        self._output_stream = collections.deque(maxlen=self.MAX_PIPE_LENGTH)
        for level in [logging.DEBUG]:
            handler = U.QueueLogger(self._output_stream)
            handler.setLevel(level)
            self._logger.addHandler(handler)

    ###########################
    ##### PRIVATE METHODS #####
    ###########################
    def _launch_minecraft(self, port, minecraft_dir, replaceable=True):
        """Launch Minecraft listening for malmoenv connections.
        Args:
            port:  the TCP port to listen on.
            installdir: the install dir name. Defaults to MalmoPlatform.
            Must be same as given (or defaulted) in download call if used.
            replaceable: whether or not to automatically restart Minecraft.
        Asserts:
            that the port specified is open.
        """
        launch_script = "launchClient.sh"
        if os.name == "nt":
            raise ValueError("TODO")

        launch_script = os.path.join(minecraft_dir, launch_script)
        rundir = os.path.join(minecraft_dir, "run")

        cmd = [launch_script, "-port", str(port), "-env", "-runDir", rundir]
        if self._seed:
            cmd += ["-seed", str(self._seed)]

        # add jdwp port if any set
        if self.jdwp_port:
            cmd += ["-jvm_debug_port", str(self.jdwp_port)]

        self._logger.info("Starting Minecraft process: " + str(cmd))

        if replaceable:
            cmd.append("-replaceable")
        preexec_fn = (
            os.setsid
            if "linux" in str(sys.platform) or sys.platform == "darwin"
            else None
        )
        minecraft_process = psutil.Popen(
            cmd,
            cwd=MC_DIR,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            # use process group, see http://stackoverflow.com/a/4791612/18576
            preexec_fn=preexec_fn,
        )
        return minecraft_process

    @staticmethod
    def _kill_minecraft_via_malmoenv(host, port):
        """Use carefully to cause the Minecraft service to exit (and hopefully restart)."""
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(1)
            sock.connect((host, port))
            U.send_message(sock, ("<MalmoEnv" + MALMO_VERSION + "/>").encode())

            U.send_message(sock, ("<Exit>NOW</Exit>").encode())
            reply = U.recv_message(sock)
            (ok,) = struct.unpack("!I", reply)
            sock.close()
            return ok == 1
        except Exception as e:
            logger.error(
                f"Attempted to send kill command to minecraft process and failed with exception {e}"
            )
            return False

    def __del__(self):
        """
        On destruction of this instance kill the child.
        """
        self._destruct()

    def _destruct(self, should_close=False):
        """
        Do our best as the parent process to destruct and kill the child + watcher.
        """
        from ..utils import watchdog
        from .manager import InstanceManager

        if (self.running or should_close) and not self.existing:
            self.running = False
            self._starting = False

            # Wait for the process to start.
            time.sleep(1)
            if self._kill_minecraft_via_malmoenv(self.host, self.port):
                # Let the minecraft process term on its own terms.
                time.sleep(2)

            # Now lets try and end the process if anything is lying around
            watchdog.reap_process_and_children(self.minecraft_process)

            # kill the minecraft process and its subprocesses
            try:
                shutil.rmtree(self.instance_dir)
            except:
                print("Failed to delete the temporary minecraft directory.")

            if self in InstanceManager._instance_pool:
                InstanceManager._instance_pool.remove(self)
                self.release_lock()
        pass

    def __repr__(self):
        return f"Malmo[{self.role}:{self.uuid}, proc={self.minecraft_process.pid if not self.existing else 'EXISTING'}, addr={self.host}:{self.port}, locked={self.locked}]"

    def _acquire_lock(self, owner=None):
        self.locked = True
        self.owner = owner

    def release_lock(self):
        self.locked = False
        self.owner = None

    def _log_heuristic(self, msg):
        """
        Log the message, heuristically determine logging level based on the
        message content
        """
        if (
            "STDERR" in msg
            or "ERROR" in msg
            or "Exception" in msg
            or "    at " in msg
            or msg.startswith("Error")
        ) and ("connection closed, likely by peer" not in msg):
            self._logger.error(msg)
        elif "WARN" in msg:
            self._logger.warning(msg)
        elif "LOGTOPY" in msg:
            self._logger.info(msg)
        else:
            self._logger.debug(msg)



from __future__ import print_function

# ------------------------------------------------------------------------------------------------
# Copyright (c) 2016 Microsoft Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
# associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
# NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
# ------------------------------------------------------------------------------------------------

# Used for integration tests.

from builtins import str
from builtins import range
import io
import os
import platform
import socket
import subprocess
import sys
import time


def _port_has_listener(port):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    result = sock.connect_ex(("127.0.0.1", port))
    sock.close()
    return result == 0


def launch_minecraft_in_background(
    minecraft_path, ports=None, timeout=360, replaceable=False, score=False
):
    if ports is None:
        ports = []
    if len(ports) == 0:
        ports = [10000]  # Default
    processes = []
    for port in ports:
        if _port_has_listener(port):
            print(
                "Something is listening on port",
                port,
                "- will assume Minecraft is running.",
            )
            continue
        replaceable_arg = " -replaceable " if replaceable else ""
        scorepolicy_arg = " -scorepolicy " if score else ""
        scorepolicy_value = " 2 " if score else ""
        print(
            "Nothing is listening on port",
            port,
            "- will attempt to launch Minecraft from a new terminal.",
        )
        if os.name == "nt":
            args = [
                minecraft_path + "/launchClient.bat",
                "-port",
                str(port),
                replaceable_arg.strip(),
                scorepolicy_arg.strip(),
                scorepolicy_value.strip(),
            ]
            p = subprocess.Popen(
                [arg for arg in args if arg != ""],
                creationflags=subprocess.CREATE_NEW_CONSOLE,
                close_fds=True,
            )
        elif sys.platform == "darwin":
            # Can't pass parameters to launchClient via Terminal.app, so create a small launch
            # script instead.
            # (Launching a process to run the terminal app to run a small launch script to run
            # the launchClient script to run Minecraft... is it possible that this is not the most
            # straightforward way to go about things?)
            launcher_file = "/tmp/launcher_" + str(os.getpid()) + ".sh"
            tmp_file = open(launcher_file, "w")
            tmp_file.write(
                minecraft_path
                + "/launchClient.sh -port "
                + str(port)
                + replaceable_arg
                + scorepolicy_arg
                + scorepolicy_value
            )
            tmp_file.close()
            os.chmod(launcher_file, 0o700)
            p = subprocess.Popen(["open", "-a", "Terminal.app", launcher_file])
        else:
            p = subprocess.Popen(
                minecraft_path
                + "/launchClient.sh -port "
                + str(port)
                + replaceable_arg
                + scorepolicy_arg
                + scorepolicy_value,
                close_fds=True,
                shell=True,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
        processes.append(p)
        print("Giving Minecraft some time to launch... ")
        launched = False
        for _ in range(timeout // 3):
            print(".", end=" ")
            time.sleep(3)
            if _port_has_listener(port):
                print("ok")
                launched = True
                break
        if not launched:
            print("Minecraft not yet launched. Giving up.")
            exit(1)
    return processes


if __name__ == "__main__":
    minecraft_launch_path = os.path.dirname(os.path.abspath(__file__))
    launch_ports = [
        int(port_arg)
        for port_arg in sys.argv[1:]
        if port_arg != "--replaceable" and port_arg != "--score"
    ]
    launch_minecraft_in_background(
        minecraft_launch_path,
        launch_ports,
        300,
        replaceable="--replaceable" in sys.argv,
        score="--score" in sys.argv,
    )



from minedojo.sim.handlers.agent import Action
from minedojo.sim.spaces import DiscreteRange, Dict


class SwapSlotAction(Action):
    """
    An action handler for swapping items in two slots
    """

    xml_command = "swapInventoryItems"

    def __init__(self, min_slot_idx: int = 0, max_slot_idx: int = 40):
        space = Dict(
            {
                "source_slot": DiscreteRange(begin=min_slot_idx, end=max_slot_idx + 1),
                "target_slot": DiscreteRange(begin=min_slot_idx, end=max_slot_idx + 1),
            }
        )
        super().__init__(command="swap_slot", space=space)

    def xml_template(self) -> str:
        return str("<InventoryCommands/>")

    def to_hero(self, x):
        return f'{self.xml_command} {x["source_slot"]} {x["target_slot"]}'



import os
import json
import time
import struct
import socket
import logging
from concurrent.futures import ThreadPoolExecutor
from typing import Optional, List, NamedTuple, Dict

import numpy as np
from lxml import etree

from ..mc_instance import InstanceManager, MinecraftInstance
from ..utils import retry


MALMO_VERSION = "0.37.0"
# Time to wait before raising an exception (high value because some operations we wait on are very slow)
MAX_WAIT = 600

logger = logging.getLogger(__name__)


class StepTuple(NamedTuple):
    """
    step_success: whether this MC client step is successful
    raw_obs: if step successful, will be a dict (key is player index) of raw obs dict. Otherwise, will be none
    """

    step_success: bool
    raw_obs: Optional[Dict[int, dict]]


class BridgeEnv:
    """
    Bridge MineDojo sim and MineCraft java world.
    Not necessarily need to be a gym.env sub-class.
    """

    MALMO_VERSION = MALMO_VERSION
    # specifies if turnkey and info are included in message.
    STEP_OPTIONS = 0
    # After this much time a socket exception will be thrown.
    SOCKTIME = 60.0 * 4

    def __init__(
        self,
        *,
        agent_count: int = 1,
        is_fault_tolerant: bool = True,
        seed: Optional[int] = None,
    ):
        assert agent_count == 1, "TODO"
        self._agent_count = agent_count
        self._rng = np.random.default_rng(seed=seed)
        self._instances: List[MinecraftInstance] = []
        self._is_fault_tolerant = is_fault_tolerant
        self._already_closed = False
        self._terminated = False

        self._seed_instance_manager()

    @property
    def is_terminated(self):
        return self._terminated

    def reset(self, episode_uid: str, agent_xmls: List[etree.Element]):
        # seed the manager
        self._seed_instance_manager()

        # Start missing instances, quit episodes, and make socket connections
        self._setup_instances()
        self._terminated = False

        # Start the Mission/Task by sending the master mission XML over
        # the pipe to these instances, and  update the agent xmls to get
        # the port/ip of the master agent send the remaining XMLS.
        self._send_mission(
            self._instances[0], agent_xmls[0], self._get_token(0, episode_uid)
        )  # Master
        if self._agent_count > 1:
            raise ValueError("TODO")

        return self._query_first_obs()

    def step(self, action_xmls: List[str]):
        """
        Bridge the MC java client step.

        Args:
            action_xmls: A list of prepared action XMLs.
        """
        assert len(action_xmls) == len(
            self._instances
        ), f"Expect {len(self._instances)} action XMLs, received {len(action_xmls)} instead"
        if not self._terminated:
            assert self.STEP_OPTIONS in {0, 2}
            all_obs = {}
            any_done = False
            for i, instance in enumerate(self._instances):
                try:
                    malmo_command = action_xmls[i]
                    step_message = f"<StepClient{str(self.STEP_OPTIONS)}>{malmo_command}</StepClient{str(self.STEP_OPTIONS)} >"
                    # Send Actions.
                    instance.client_socket_send_message(step_message.encode())
                    # Receive the (image) observation.
                    obs = instance.client_socket_recv_message()
                    # Receive reward (useless though), done, and sent.
                    reply = instance.client_socket_recv_message()
                    _, done, sent = struct.unpack("!dbb", reply)
                    any_done = any_done or (done == 1)
                    # Receive info from the environment.
                    malmo_json = instance.client_socket_recv_message().decode("utf-8")
                    raw = json.loads(malmo_json) if malmo_json is not None else {}
                    raw["pov"] = obs
                    all_obs[i] = raw
                except (socket.timeout, socket.error, TypeError) as e:
                    # when the socket times out...
                    self._terminated = True
                    logger.error(f"Failed to take a step. Error msg: {e}")
                    return StepTuple(step_success=False, raw_obs=None)
            self._terminated = any_done

            # step the server
            # instance[0] is the server
            server = self._instances[0]
            step_message = "<StepServer></StepServer>"
            try:
                server.client_socket_send_message(step_message.encode())
            except (socket.timeout, socket.error, TypeError) as e:
                self._terminated = True
                logger.error("Failed to take a step (timeout or error).")
        else:
            raise RuntimeError("Attempted to step an environment server with done=True")
        return StepTuple(step_success=True, raw_obs=all_obs)

    def close(self):
        logger.debug("Closing...")
        if self._already_closed:
            return
        for instance in self._instances:
            self._clean_connection(instance)
            if instance.running:
                instance.kill()
        self._already_closed = True

    def _setup_instances(self):
        """
        Set up MC instances
        """
        n_instances_to_start = self._agent_count - len(self._instances)
        if n_instances_to_start > 0:
            instance_futures = []
            with ThreadPoolExecutor(max_workers=n_instances_to_start) as tpe:
                for _ in range(n_instances_to_start):
                    instance_futures.append(tpe.submit(self._get_new_instance))
            self._instances.extend([f.result() for f in instance_futures])

        # establish socket connections
        for instance in reversed(self._instances):
            self._clean_connection(instance)
            self._create_connection(instance)
            # TODO: Properly rewrite fault tolerance.
            self._quit_current_episode(instance)

    def _get_new_instance(self, port=None, instance_id=None) -> MinecraftInstance:
        """
        Gets a new instance and sets up a logger if need be.
        """
        if port is not None:
            instance = InstanceManager.add_existing_instance(port)
        else:
            instance = InstanceManager.get_instance(
                os.getpid(), instance_id=instance_id
            )

        instance.launch(replaceable=self._is_fault_tolerant)
        instance.had_to_clean = False
        return instance

    def _query_first_obs(self):
        all_obs = {}
        if not self._terminated:
            logger.debug("Query the first obs.")
            peek_message = "<Peek/>"
            any_done = False
            for i, instance in enumerate(self._instances):
                st_time = time.time()
                instance.client_socket_send_message(peek_message.encode())
                obs = instance.client_socket_recv_message()
                info = instance.client_socket_recv_message().decode("utf-8")

                reply = instance.client_socket_recv_message()
                (done,) = struct.unpack("!b", reply)
                any_done = any_done or (done == 1)
                if obs is None or len(obs) == 0:
                    if time.time() - st_time > MAX_WAIT:
                        instance.client_socket_close()
                        raise Exception("too long waiting for first observation")
                    time.sleep(0.1)
                    # FIXME - shouldn't we error or retry here?

                raw = json.loads(info) if info is not None else {}
                raw["pov"] = obs
                all_obs[i] = raw
            self._terminated = any_done
            if self._terminated:
                raise RuntimeError(
                    "Something went wrong resetting the environment! "
                    "`done` was true on first frame."
                )
        return all_obs

    @retry
    def _create_connection(self, instance: MinecraftInstance):
        try:
            logger.debug(f"Creating socket connection {instance}")
            instance.create_instance_socket(socktime=self.SOCKTIME)
            logger.debug(f"Saying hello for client: {instance}")
            self._hello_server(instance)
        except (socket.timeout, socket.error, ConnectionRefusedError) as e:
            instance.had_to_clean = True
            logger.error("Failed to reset (socket error), trying again!")
            logger.error("Cleaning connection! Something must have gone wrong.")
            self._clean_connection(instance)
            self._kill_frozen_instance(instance)
            raise e

    def _seed_instance_manager(self):
        InstanceManager.seed_manager(self._rng.integers(low=0, high=2**31 - 1))

    @staticmethod
    def _send_mission(
        instance: MinecraftInstance,
        mission_xml_etree: etree.Element,
        token_in: str,
        agent_count: int = 1,
        seed: Optional[int] = None,
    ):
        """
        Send the mission XML to the given isntance.
        """
        # init all instance missions
        ok = 0
        st_time = time.time()
        logger.debug(f"Sending mission init: {instance}")
        while ok != 1:
            # roundtrip through etree to escape symbols correctly and make printing pretty
            mission_xml = etree.tostring(mission_xml_etree)
            token = f"{token_in}:{str(agent_count)}:true"
            if seed is not None:
                token += f":{seed}"
            token = token.encode()
            instance.client_socket_send_message(mission_xml)
            instance.client_socket_send_message(token)

            reply = instance.client_socket_recv_message()
            (ok,) = struct.unpack("!I", reply)
            if ok != 1:
                if (time.time() - st_time) > MAX_WAIT:
                    raise socket.timeout()
                logger.debug(f"Recieved a MALMOBUSY from {instance}; trying again.")
                time.sleep(1)

    @staticmethod
    def _clean_connection(instance: MinecraftInstance):
        """
        Clean the connection given an instance.
        No effects at the first time.
        """
        try:
            if instance.has_client_socket():
                # Try to disconnect gracefully.
                try:
                    instance.client_socket_send_message("<Disconnect/>".encode())
                except:
                    pass
                instance.client_socket_shutdown(socket.SHUT_RDWR)
        except (BrokenPipeError, OSError, socket.error):
            # There is no connection left!
            instance.client_socket = None

    @staticmethod
    def _hello_server(instance: MinecraftInstance):
        instance.client_socket_send_message(
            ("<MalmoEnv" + MALMO_VERSION + "/>").encode()
        )

    @staticmethod
    def _kill_frozen_instance(instance: MinecraftInstance):
        if instance.had_to_clean:
            logger.error(
                f"Connection with Minecraft client {instance} cleaned more than once; restarting."
            )

            instance.kill()
        else:
            instance.had_to_clean = True

    @staticmethod
    def _quit_current_episode(instance: MinecraftInstance):
        has_quit = False

        logger.info(f"Attempting to quit: {instance}")
        # while not has_quit:
        instance.client_socket_send_message("<Quit/>".encode())
        reply = instance.client_socket_recv_message()
        (ok,) = struct.unpack("!I", reply)
        has_quit = not (ok == 0)
        # TODO: Get this to work properly

        # time.sleep(0.1)

    @staticmethod
    def _get_token(role: int, ep_uid: str):
        return f"{ep_uid}:{str(role)}:0"



import os
import sys
import zipfile

import requests
import wget


def get_doi(link):
    response = requests.get(link)
    return response.url


DOWNLOAD_URLS = {
    "youtube": {
        "full": f"{get_doi('https://doi.org/10.5281/zenodo.6641142')}/files/youtube_full.json",
        "tutorial": f"{get_doi('https://doi.org/10.5281/zenodo.6641142')}/files/youtube_tutorial.json",
    },
    "wiki": {
        "full": f"{get_doi('https://doi.org/10.5281/zenodo.6640448')}/files/wiki_full.zip",
        "samples": f"{get_doi('https://doi.org/10.5281/zenodo.6640448')}/files/wiki_samples.zip",
    },
    "reddit": {
        "full": f"{get_doi('https://doi.org/10.5281/zenodo.6641114')}/files/reddit.json",
    },
}


def get_fn(source: str = "youtube", download_dir: str = None, full: bool = True):
    url = DOWNLOAD_URLS[source][
        "full" if full else "tutorial" if source == "youtube" else "samples"
    ]
    fn = wget.filename_from_url(url)
    extracted_fn = fn if source != "wiki" else fn.replace(".zip", "")
    extracted_fn = os.path.join(download_dir, extracted_fn)
    download_fn = os.path.join(download_dir, fn)

    return extracted_fn, download_fn, url


def download(source: str = "youtube", download_dir: str = None, full: bool = True):
    r"""Function that downloads the datasets of MineDojo to local.

    Args:
        source (string): the sub-dataset name:
            ``'youtube'`` | ``'wiki'`` | ``'reddit'``
        download_dir (string, optional): Directory path where the downloaded data will be saved.
            The default directory is ``~/.minedojo/``.
        full (bool, optional): Only useful when ``source='youtube'`` or ``source='wiki'``. If set to ``False``, only the tutorial version
            of YouTube dataset or the sample version of Wiki dataset will be downloaded. Default is ``True``.

    Returns:
        Directory path to downloaded (extracted) file or folder.

    Examples::
        >>> from minedojo.data.download import download
        >>> print(download("youtube", full=False))
        '~/.minedojo/youtube_tutorial.json'
        >>> print(download("wiki", full=False))
        '~/.minedojo/wiki_samples'
        >>> print(download("reddit"))
        '~/.minedojo/reddit.json'
    """

    if download_dir is None:
        download_dir = os.path.join(os.path.expanduser("~"), ".minedojo")
    if not os.path.exists(download_dir):
        os.makedirs(download_dir)

    if source == "reddit" and not full:
        print(
            "Reddit dataset does not have a sample version. Will download / use the full version."
        )
        full = True

    extracted_fn, download_fn, url = get_fn(source, download_dir, full)

    if os.path.exists(extracted_fn) and (
        source != "wiki" or not os.path.exists(download_fn)
    ):
        return extracted_fn
    elif not os.path.exists(download_fn):
        print(f"Downloading {url} to {download_fn}...")
        wget.download(url, out=download_fn, bar=bar_progress)

    if source == "wiki":
        with zipfile.ZipFile(download_fn, "r") as zip_ref:
            zip_ref.extractall(download_dir)
        os.remove(download_fn)

    return extracted_fn


def bar_progress(current, total, width=80):
    progress_message = f"Downloading: {current / total * 100:.1f}% [{current/1e6:.1f} / {total/1e6:.1f}] MB"
    sys.stdout.write("\r" + progress_message)
    sys.stdout.flush()



import uuid
from copy import deepcopy
from typing import Union, Optional, List, Dict, Tuple, Literal, Any

import cv2
import gym
import numpy as np
from lxml import etree

from .mc_meta import mc
from . import handlers
from .bridge import BridgeEnv
from .cmd_executor import CMDExecutor
from .config_sim_spec import SimSpec
from .inventory import InventoryItem, parse_inventory_item


class MineDojoSim(gym.Env):
    """An environment wrapper for MineDojo simulation.

    Args:
        allow_mob_spawn: If ``True``, allow mobs (animals and hostiles) to spawn.
                Default: ``True``.

        allow_time_passage: Time flows if ``True``.
                Default: ``True``.

        break_speed_multiplier: Controls the speed of breaking blocks. A value larger than 1.0 accelerates the breaking.
                Default: ``1.0``.

        drawing_str: Draws shapes (e.g. spheres, cuboids) in the minecraft world.
                Default: ``None``.

        event_level_control: If ``True``, the agent is able to perform high-level controls including place and equip.
                If ``False``, then is keyboard-mouse level control.
                Default: ``True``.

        flat_world_seed_string: A string that specifies how we want the world layers to be created
                if ``generate_world_type`` is "flat".
                One can use the `tool <https://minecraft.tools/en/flat.php?biome=1&bloc_1_nb=1&bloc_1_id=2&bloc_2_nb=2&bloc_2_id=3%2F00&bloc_3_nb=1&bloc_3_id=7&village_size=1&village_distance=32&mineshaft_chance=1&stronghold_count=3&stronghold_distance=32&stronghold_spread=3&oceanmonument_spacing=32&oceanmonument_separation=5&biome_1_distance=32&valid=Create+the+Preset#seed>`_ to generate.
                Default: ``None``.

        generate_world_type: A string that specifies the type of the minecraft world.
                One of ``"default"``, ``"flat"``, ``"from_file"``, ``"specified_biome"``.
                Default: ``"default"``.

        image_size: The size of image observations.

        initial_inventory: If not ``None``, specifies initial items in the agent's inventory.
                Use class ``InventoryItem`` to specify items.
                Default: ``None``.

        initial_weather: If not ``None``, specifies the initial weather.
                Can be one of ``"clear"``, ``"normal"``, ``"rain"``, ``"thunder"``.
                Default: ``None``.

        lidar_rays: Defines the directions and maximum distances of the lidar rays if ``use_lidar`` is ``True``.
                If supplied, should be a list of tuple(pitch, yaw, distance).
                Pitch and yaw are in radians and relative to agent looking vector.
                Default: ``None``.

        raise_error_on_invalid_cmds: If ``True``, the cmd executor will raise error when a command is invalid.
                If ``False``, the executor will just skip instead.
                Default: ``False``.

        regenerate_world_after_reset: If ``True``, the minecraft world will be re-generated when resetting.
                Default: ``False``.

        seed: The seed for an instance's internal random number generator.
                Default: ``None``.

        sim_name: Name of a simulation instance.
                Default: ``"MineDojoSim"``.

        spawn_in_village: If ``True``, the agent will spawn in a village.
                Default: ``False``.

        specified_biome: If not ``None``, specifies the biome of the minecraft world by a string or an integer.
                Default: ``None``.

        start_food: If not ``None``, specifies initial food of the agent.
                Default: ``None``.

        start_health: If not ``None``, specifies initial health of the agent.
                Default: ``None``.

        start_position: If not ``None``, specifies the agent's initial location and orientation.
                If provided, should be a dict with keys ``x``, ``y``, ``z``, ``yaw``, ``pitch``.
                ``yaw`` and ``pitch`` are in degrees.
                Default: ``None``.

        start_time: If not ``None``, specifies the time when the agent spawns.
                If supplied, should be an int between 0 and 24000.
                See `here <https://minecraft.fandom.com/wiki/Daylight_cycle>`_ for more information.
                Default: ``None``.

        use_depth: If ``True``, includes depth map in observations.
                Default: ``False``.

        use_lidar: If ``True``, includes lidar in observations.
                Default: ``False``.

        use_voxel: If ``True``, includes voxel in observations.
                Default: ``False``.

        voxel_size: Defines the voxel's range in each axis if ``use_voxel`` is ``True``.
                If supplied, should be a dict with keys ``xmin``, ``xmax``, ``ymin``, ``ymax``, ``zmin``, ``zmax``.
                Each value specifies the voxel size relative to the agent.
                Default: ``None``.

        world_file_path: The path to the world file if ``generate_world_type`` is ``"from_file"``.
                Default: ``None``.

        world_seed: Seed for deterministic world generation
                if ``generate_world_type`` is ``"default"`` or ``"specified_biome"``.
                See `here <https://minecraft.fandom.com/wiki/Seed_(level_generation)>`_ for more details.
                Default: ``None``.
    """

    def __init__(
        self,
        *,
        # ------ initial conditions ------
        initial_inventory: Optional[List[InventoryItem]] = None,
        start_position: Optional[Dict[str, Union[float, int]]] = None,
        start_health: Optional[float] = None,
        start_food: Optional[int] = None,
        start_time: Optional[int] = None,
        initial_weather: Optional[Literal["normal", "clear", "rain", "thunder"]] = None,
        spawn_in_village: bool = False,
        drawing_str: Optional[str] = None,
        # ------ global conditions ------
        break_speed_multiplier: float = 1.0,
        allow_time_passage: bool = True,
        allow_mob_spawn: bool = True,
        # ------ world generation ------
        generate_world_type: Literal[
            "default", "from_file", "flat", "specified_biome"
        ] = "default",
        regenerate_world_after_reset: bool = False,
        world_seed: Optional[Union[str, int]] = None,
        world_file_path: Optional[str] = None,
        flat_world_seed_string: Optional[str] = None,
        specified_biome: Optional[Union[int, str]] = None,
        # ------ observation ------
        image_size: Union[int, Tuple[int, int]],
        use_voxel: bool = False,
        voxel_size: Optional[Dict[str, int]] = None,
        use_lidar: bool = False,
        lidar_rays: Optional[List[Tuple[float, float, float]]] = None,
        use_depth: bool = False,
        # ------ control ------
        event_level_control: bool = True,
        # ------ randomness ------
        seed: Optional[int] = None,
        # ------ misc ------
        sim_name: str = "MineDojoSim",
        raise_error_on_invalid_cmds: bool = False,
    ):
        self._sim_name = sim_name
        self._rng = np.random.default_rng(seed)
        if isinstance(image_size, int):
            image_size = (image_size, image_size)
        if use_voxel:
            voxel_size = voxel_size or dict(
                xmin=-1, ymin=-1, zmin=-1, xmax=1, ymax=1, zmax=1
            )
        if use_lidar:
            lidar_rays = lidar_rays or [(0.0, 0.0, 10.0)]
        assert not use_depth, "TODO: fix depth obs bug"
        start_health = start_health or 20.0
        start_food = start_food or 20
        assert generate_world_type in {
            "default",
            "from_file",
            "flat",
            "specified_biome",
        }, f"invalid generate world type {generate_world_type}"
        if generate_world_type == "default":
            world_seed = world_seed or ""
        elif generate_world_type == "from_file":
            assert (
                world_file_path is not None
            ), f'A world file path must be provided when `generate_world_type = "from_file"`'
        elif generate_world_type == "flat":
            flat_world_seed_string = flat_world_seed_string or ""
        elif generate_world_type == "specified_biome":
            assert (
                specified_biome is not None
            ), f"must provide a biome when generate_world_type = 'specified_biome'!"
            world_seed = world_seed or ""
        else:
            raise ValueError(f"Unknown world generation type {generate_world_type}")
        if specified_biome is not None:
            if isinstance(specified_biome, str):
                assert (
                    specified_biome in mc.BIOMES_MAP
                ), f"Unknown biome name {specified_biome}"
                specified_biome = mc.BIOMES_MAP[specified_biome]
            elif isinstance(specified_biome, int):
                assert (
                    specified_biome in mc.BIOMES_MAP.values()
                ), f"Invalid biome id {specified_biome}"
            else:
                raise ValueError(f"invalid biome type {specified_biome}")

        # configure obs handlers
        obs_handlers = [
            handlers.POVObservation(image_size, False),
            handlers.TrueFlatInventoryObservation(),
            handlers.EquipmentObservation(),
            handlers.ObservationFromLifeStats(),
            handlers.ObservationFromCurrentLocation(),
            handlers.ObserveFromFullStats(),
            handlers.NearbyToolsObservation(),
            handlers.ObservationFromDamageSource(),
        ]
        if use_voxel:
            voxel_size = (
                (voxel_size["xmin"], voxel_size["xmax"]),
                (voxel_size["ymin"], voxel_size["ymax"]),
                (voxel_size["zmin"], voxel_size["zmax"]),
            )
            obs_handlers.append(handlers.VoxelObservation(voxel_size))
        if use_lidar:
            obs_handlers.append(handlers.RichLidarObservation(lidar_rays))
        # configure action handlers
        common_actions = [
            "forward",
            "back",
            "left",
            "right",
            "jump",
            "sneak",
            "sprint",
            "use",
            "attack",
            "drop",
        ]
        action_handlers = [
            handlers.CameraAction(),
            handlers.SmeltAction(
                ["none"] + mc.ALL_SMELTING_ITEMS, _other="none", _default="none"
            ),
            handlers.CraftAction(
                ["none"] + mc.ALL_PERSONAL_CRAFTING_ITEMS,
                _other="none",
                _default="none",
            ),
            handlers.CraftWithTableAction(
                ["none"] + mc.ALL_CRAFTING_TABLE_ITEMS, _other="none", _default="none"
            ),
        ]
        action_handlers.extend(
            [
                handlers.KeybasedCommandAction(k, mc.INVERSE_KEYMAP[k])
                for k in common_actions
            ]
        )
        if event_level_control:
            action_handlers.extend(
                [
                    handlers.EquipAction(
                        ["none"] + mc.ALL_ITEMS, _other="none", _default="none"
                    ),
                    handlers.PlaceBlock(
                        ["none"] + mc.ALL_ITEMS, _other="none", _default="none"
                    ),
                ]
            )
        else:
            action_handlers.append(handlers.SwapSlotAction())
            action_handlers.append(
                handlers.KeybasedCommandAction(
                    "pickItem", mc.INVERSE_KEYMAP["pickItem"]
                )
            )
            action_handlers.extend(
                [
                    handlers.KeybasedCommandAction(
                        f"hotbar.{i}", mc.INVERSE_KEYMAP[str(i)]
                    )
                    for i in range(1, 10)
                ]
            )
        # configure agent handlers
        agent_handlers = []
        # configure agent start handlers, e.g., initial inventory
        self.start_health, self.start_food = start_health, start_food
        agent_start_handlers = [
            handlers.LowLevelInputsAgentStart(),
            handlers.AgentStartBreakSpeedMultiplier(break_speed_multiplier),
            handlers.StartingHealthAgentStart(health=start_health),
            handlers.StartingFoodAgentStart(food=start_food),
        ]
        self.initial_inventory = initial_inventory
        if initial_inventory is not None:
            initial_inventory = [
                parse_inventory_item(item) for item in initial_inventory
            ]
            agent_start_handlers.append(
                handlers.InventoryAgentStart(
                    {
                        inventory_item[0]: inventory_item[1]
                        for inventory_item in initial_inventory
                    }
                )
            )
        self.start_position = start_position
        if start_position is not None:
            agent_start_handlers.append(
                handlers.AgentStartPlacement(
                    x=start_position["x"],
                    y=start_position["y"],
                    z=start_position["z"],
                    yaw=start_position["yaw"],
                    pitch=start_position["pitch"],
                )
            )
        # configure server initial conditions handlers
        self.start_time = start_time
        server_initial_conditions_handlers = [
            handlers.TimeInitialCondition(
                allow_passage_of_time=allow_time_passage, start_time=start_time
            ),
            handlers.SpawningInitialCondition(allow_mob_spawn),
        ]
        self.initial_weather = initial_weather
        if initial_weather is not None:
            server_initial_conditions_handlers.append(
                handlers.WeatherInitialCondition(initial_weather)
            )
        # configure world generator handlers
        world_generator_handlers = []
        if generate_world_type == "default":
            world_generator_handlers.append(
                handlers.DefaultWorldGenerator(regenerate_world_after_reset, world_seed)
            )
        elif generate_world_type == "from_file":
            world_generator_handlers.append(
                handlers.FileWorldGenerator(
                    world_file_path,
                    destroy_after_use=False,
                )
            )
        elif generate_world_type == "flat":
            world_generator_handlers.append(
                handlers.FlatWorldGenerator(
                    force_reset=True, generatorString=flat_world_seed_string
                )
            )
        elif generate_world_type == "specified_biome":
            world_generator_handlers.append(
                handlers.BiomeGenerator(
                    specified_biome, force_reset=True, world_seed=world_seed
                )
            )
        else:
            raise ValueError()
        if drawing_str is not None:
            world_generator_handlers.append(handlers.DrawingDecorator(drawing_str))
        # configure server decorator handlers
        server_decorator_handlers = []
        if spawn_in_village:
            server_decorator_handlers.append(handlers.VillageSpawnDecorator())
        # configure server quit handlers
        server_quit_handlers = []
        self._sim_spec = SimSpec(
            sim_name=sim_name,
            agent_count=1,
            obs_handlers=obs_handlers,
            action_handlers=action_handlers,
            agent_handlers=agent_handlers,
            agent_start_handlers=agent_start_handlers,
            server_initial_conditions_handlers=server_initial_conditions_handlers,
            world_generator_handlers=world_generator_handlers,
            server_decorator_handlers=server_decorator_handlers,
            server_quit_handlers=server_quit_handlers,
            seed=self.new_seed,
        )

        self._bridge_env = BridgeEnv(is_fault_tolerant=True, seed=self.new_seed)

        self._prev_obs = None
        self._prev_action = None
        self._prev_info = None

        self._cmd_executor = CMDExecutor(self, raise_error_on_invalid_cmds)

    @property
    def observation_space(self):
        return self._sim_spec.observation_space

    @property
    def action_space(self):
        return self._sim_spec.action_space

    @property
    def new_seed(self):
        return self._rng.integers(low=0, high=2**31 - 1).item()

    def seed(self, seed: int = None):
        """Sets the seed for this env’s random number generator.

        Args:
            seed: The seed for the number generator
        """
        self._rng = np.random.default_rng(seed)

    def reset(self):
        """Resets the environment to an initial state and returns an initial observation.

        Return:
            Agent’s initial observation.
        """
        episode_id = str(uuid.uuid4())

        xml = etree.fromstring(self._sim_spec.to_xml(episode_id))
        raw_obs = self._bridge_env.reset(episode_id, [xml])[0]
        obs, info = self._process_raw_obs(raw_obs)
        self._prev_obs, self._prev_info = deepcopy(obs), deepcopy(info)
        return obs

    def step(self, action: dict):
        """Run one timestep of the environment’s dynamics. Accepts an action and returns next_obs, reward, done, info.

        Args:
            action: The action of the agent in current step.

        Return:
            A tuple (obs, reward, done, info)
            - ``dict`` - Agent’s next observation.
            - ``float`` - Amount of reward returned after executing previous action.
            - ``bool`` - Whether the episode has ended.
            - ``dict`` - Contains auxiliary diagnostic information (helpful for debugging, and sometimes learning).
        """
        self._prev_action = deepcopy(action)
        action_xml = self._action_obj_to_xml(action)
        step_tuple = self._bridge_env.step([action_xml])
        step_success, raw_obs = step_tuple.step_success, step_tuple.raw_obs
        if not step_success:
            # when step failed, return prev obs
            return self._prev_obs, 0, True, self._prev_info
        else:
            obs, info = self._process_raw_obs(raw_obs[0])
            self._prev_obs, self._prev_info = deepcopy(obs), deepcopy(info)
            return obs, 0, self.is_terminated, info

    def execute_cmd(self, cmd: str, action: Optional[dict] = None):
        """Execute a given string command.

        Args:
            cmd: The string command accepted by the Minecraft client.
            action: An action that will be simultaneously executed with the command.

        Return:
            A tuple (obs, reward, done, info)
            - ``dict`` - Agent’s observation of the current environment.
            - ``float`` - Amount of reward returned after previous action.
            - ``bool`` - Whether the episode has ended.
            - ``dict`` - Contains auxiliary diagnostic information (helpful for debugging, and sometimes learning).
        """
        return self._cmd_executor.execute_cmd(cmd, action)

    def spawn_mobs(
        self,
        mobs: Union[str, List[str]],
        rel_positions: Union[np.ndarray, list],
        action: Optional[dict] = None,
    ):
        """Spawn mobs in the world.

        Args:
            mobs: The names of the mobs to spawn
            rel_positions: The mobs' positions relative to the agent
            action: An action that will be simultaneously executed with the spawning
        Return:
            A tuple (obs, reward, done, info)
            - ``dict`` - Agent’s observation of the current environment.
            - ``float`` - Amount of reward returned after previous action.
            - ``bool`` - Whether the episode has ended.
            - ``dict`` - Contains auxiliary diagnostic information (helpful for debugging, and sometimes learning).
        """
        return self._cmd_executor.spawn_mobs(mobs, rel_positions, action)

    def set_block(
        self,
        blocks: Union[str, List[str]],
        rel_positions: Union[np.ndarray, list],
        action: Optional[dict] = None,
    ):
        """Set blocks in the world.

        Args:
            blocks: The names of the blocks to set
            rel_positions: The blocks' positions relative to the agent
            action: An action that will be simultaneously executed with the setting
        Return:
            A tuple (obs, reward, done, info)
            - ``dict`` - Agent’s observation of the current environment.
            - ``float`` - Amount of reward returned after previous action.
            - ``bool`` - Whether the episode has ended.
            - ``dict`` - Contains auxiliary diagnostic information (helpful for debugging, and sometimes learning).
        """
        return self._cmd_executor.set_block(blocks, rel_positions, action)

    def clear_inventory(self, action: Optional[dict] = None):
        """Remove all items in the agent's inventory.

        Args:
            action: An action that will be simultaneously executed
        Return:
            A tuple (obs, reward, done, info)
            - ``dict`` - Agent’s observation of the current environment.
            - ``float`` - Amount of reward returned after previous action.
            - ``bool`` - Whether the episode has ended.
            - ``dict`` - Contains auxiliary diagnostic information (helpful for debugging, and sometimes learning).
        """
        return self._cmd_executor.clear_inventory(action)

    def set_inventory(
        self, inventory_list: List[InventoryItem], action: Optional[dict] = None
    ):
        """Set items to the agent's inventory.

        Args:
            inventory_list: List of ``InventoryItem`` to change the inventory status
            action: An action that will be simultaneously executed
        Return:
            A tuple (obs, reward, done, info)
            - ``dict`` - Agent’s observation of the current environment.
            - ``float`` - Amount of reward returned after previous action.
            - ``bool`` - Whether the episode has ended.
            - ``dict`` - Contains auxiliary diagnostic information (helpful for debugging, and sometimes learning).
        """
        return self._cmd_executor.set_inventory(inventory_list, action)

    def teleport_agent(self, x, y, z, yaw, pitch, action: Optional[dict] = None):
        """Teleport the agent to a given position.

        Args:
            x: x coordinate of the destination
            y: y coordinate of the destination
            z: z coordinate of the destination
            yaw: yaw of the targeted orientation
            pitch: pitch of the targeted orientation
            action: An action that will be simultaneously executed with the teleporting
        Return:
            A tuple (obs, reward, done, info)
            - ``dict`` - Agent’s observation of the current environment.
            - ``float`` - Amount of reward returned after previous action.
            - ``bool`` - Whether the episode has ended.
            - ``dict`` - Contains auxiliary diagnostic information (helpful for debugging, and sometimes learning).
        """
        return self._cmd_executor.teleport_agent(x, y, z, yaw, pitch, action)

    def kill_agent(self, action: Optional[dict] = None):
        """Kill the agent.

        Args:
            action: An action that will be simultaneously executed
        Return:
            A tuple (obs, reward, done, info)
            - ``dict`` - Agent’s observation of the current environment.
            - ``float`` - Amount of reward returned after previous action.
            - ``bool`` - Whether the episode has ended.
            - ``dict`` - Contains auxiliary diagnostic information (helpful for debugging, and sometimes learning).
        """
        return self._cmd_executor.kill_agent(action)

    def set_time(self, time: int, action: Optional[dict] = None):
        """Set the world with the given time.

        Args:
            time: The target time
            action: An action that will be simultaneously executed
        Return:
            A tuple (obs, reward, done, info)
            - ``dict`` - Agent’s observation of the current environment.
            - ``float`` - Amount of reward returned after previous action.
            - ``bool`` - Whether the episode has ended.
            - ``dict`` - Contains auxiliary diagnostic information (helpful for debugging, and sometimes learning).
        """
        return self._cmd_executor.set_time(time, action)

    def set_weather(self, weather: str, action: Optional[dict] = None):
        """Set the world with the given weather.

        Args:
            weather: The target weather
            action: An action that will be simultaneously executed
        Return:
            A tuple (obs, reward, done, info)
            - ``dict`` - Agent’s observation of the current environment.
            - ``float`` - Amount of reward returned after previous action.
            - ``bool`` - Whether the episode has ended.
            - ``dict`` - Contains auxiliary diagnostic information (helpful for debugging, and sometimes learning).
        """
        return self._cmd_executor.set_weather(weather, action)

    def random_teleport(self, max_range: int, action: Optional[dict] = None):
        """Teleport the agent randomly.

        Args:
            max_range: The maximum distance on each horizontal axis from the center of the area to spread targets
                       (thus, the area is square, not circular)
            action: An action that will be simultaneously executed
        Return:
            A tuple (obs, reward, done, info)
            - ``dict`` - Agent’s observation of the current environment.
            - ``float`` - Amount of reward returned after previous action.
            - ``bool`` - Whether the episode has ended.
            - ``dict`` - Contains auxiliary diagnostic information (helpful for debugging, and sometimes learning).
        """
        return self._cmd_executor.random_teleport(max_range, action)

    def close(self):
        """Environments will automatically close() themselves when garbage collected or when the program exits."""
        self._bridge_env.close()

    def render(self, mode: str = "human"):
        """Renders the environment.

        Args:
            mode: The mode to render with.
        """
        img = self._prev_obs["rgb"]
        img = img.transpose((1, 2, 0))
        img = img[:, :, ::-1]
        cv2.imshow(f"{self._sim_name}", img)
        cv2.waitKey(1)

    @property
    def prev_obs(self):
        return self._prev_obs

    @property
    def prev_info(self):
        return self._prev_info

    @property
    def prev_action(self):
        return self._prev_action

    @property
    def is_terminated(self):
        return self._bridge_env.is_terminated

    def _process_raw_obs(self, raw_obs: dict):
        info = deepcopy(raw_obs)
        if "pov" in info:
            info.pop("pov")
        obs_dict = {
            h.to_string(): h.from_hero(raw_obs) for h in self._sim_spec.observables
        }
        return obs_dict, info

    def _action_obj_to_xml(self, action):
        parsed_action = [f'chat {action["chat"]}'] if "chat" in action else []
        parsed_action.extend(
            [h.to_hero(action[h.to_string()]) for h in self._sim_spec.actionables]
        )
        return "\n".join(parsed_action)



# Copyright (c) 2020 All Rights Reserved
# Author: William H. Guss, Brandon Houghton
import abc
import string
import random
from collections import OrderedDict

import gym.spaces
import numpy as np


class MineRLSpace(abc.ABC, gym.Space):
    """
    An interface for MineRL spaces.
    """

    @property
    def flattened(self) -> gym.spaces.Box:
        if not hasattr(self, "_flattened"):
            self._flattened = self.create_flattened_space()
        return self._flattened

    @abc.abstractmethod
    def no_op(self, batch_shape=()):
        # TODO: ADD BATCH_SHAPE TO SAMPLE
        pass

    @abc.abstractmethod
    def create_flattened_space(self):
        pass

    @abc.abstractmethod
    def flat_map(self, x):
        pass

    @abc.abstractmethod
    def unmap(self, x):
        pass

    def is_flattenable(self):
        return True

    @abc.abstractmethod
    def sample(self, bdim=None):
        pass


class Tuple(gym.spaces.Tuple, MineRLSpace):
    def no_op(self):
        raise NotImplementedError()

    def create_flattened_space(self):
        raise NotImplementedError()

    def flat_map(self, x):
        raise NotImplementedError()

    def unmap(self, x):
        raise NotImplementedError()


class Box(gym.spaces.Box, MineRLSpace):
    def __init__(self, *args, normalizer_scale="linear", **kwargs):
        super(Box, self).__init__(*args, **kwargs)

        self._flat_low = self.low.flatten().astype(np.float64)
        self._flat_high = self.high.flatten().astype(np.float64)

        if normalizer_scale == "log":
            self.max_log = np.log(1 + (self._flat_high - self._flat_low))
        else:
            assert normalizer_scale == "linear", "only log and linear are supported"

        self.normalizer_scale = normalizer_scale

    CENTER = 0

    def no_op(self, batch_shape=()):
        return np.zeros(shape=list(batch_shape) + list(self.shape)).astype(self.dtype)

    def create_flattened_space(self):
        if len(self.shape) > 2:
            raise TypeError("Box spaces with 3D tensor shapes cannot be flattened.")
        else:
            return Box(low=self._flat_low, high=self._flat_high)

    def flat_map(self, x):
        if len(self.shape) > 0:
            flatx = x.reshape(
                list(x.shape[: -len(self.shape)]) + [np.prod(self.shape).astype(int)]
            )
        else:
            # assumes everything is in batch format, a scalar is already flattened and needs to be normalzied
            flatx = x.reshape(list(x.shape) + [-1])

        # TODO: CHECK IF THE SPACE IS BOUNDED! IN WHICH WE CANNOT NORMALIZEN USING HIGHS
        if self.normalizer_scale == "linear":
            return (flatx.astype(np.float64) - self._flat_low) / (
                self._flat_high - self._flat_low
            ) - Box.CENTER
        elif self.normalizer_scale == "log":
            return (
                np.log(flatx.astype(np.float64) - self._flat_low + 1) / self.max_log
                - Box.CENTER
            )

    def unmap(self, x):
        """
        Un-normalizes the flattened x to its original high and low.
        Then reshapes it back to the original shape.
        """
        low = x + Box.CENTER

        if self.normalizer_scale == "linear":
            high = low * (self._flat_high - self._flat_low) + self._flat_low
        elif self.normalizer_scale == "log":
            high = np.exp(low * self.max_log) - 1 + self._flat_low
        else:
            raise NotImplementedError(
                "Normalizer {} not implemented for Box space!".format(
                    self.normalizer_scale
                )
            )

        reshaped = high.reshape(list(x.shape[:-1]) + list(self.shape))
        if np.issubdtype(self.dtype, np.integer):
            return np.round(reshaped).astype(self.dtype)
        else:
            return reshaped.astype(self.dtype)

    def is_flattenable(self):
        return len(self.shape) <= 2

    def clip(self, x):
        # Clips the vector x between the vectors self.low and self.high.
        return np.clip(x, self.low, self.high)

    def sample(self, bs=None):
        """
        Generates a single random sample inside of the Box.

        In creating a sample of the box, each coordinate is sampled according to
        the form of the interval:

        * [a, b] : uniform distribution
        * [a, oo) : shifted exponential distribution
        * (-oo, b] : shifted negative exponential distribution
        * (-oo, oo) : normal distribution
        """

        bdim = () if bs is None else (bs,)

        high = self.high if self.dtype.kind == "f" else self.high.astype("int64") + 1
        sample = np.empty(bdim + self.shape)

        # Masking arrays which classify the coordinates according to interval
        # type
        unbounded = ~self.bounded_below & ~self.bounded_above
        upp_bounded = ~self.bounded_below & self.bounded_above
        low_bounded = self.bounded_below & ~self.bounded_above
        bounded = self.bounded_below & self.bounded_above

        # Vectorized sampling by interval type
        sample[..., unbounded] = self.np_random.normal(
            size=bdim + unbounded[unbounded].shape
        )

        sample[..., low_bounded] = (
            self.np_random.exponential(size=bdim + low_bounded[low_bounded].shape)
            + self.low[low_bounded]
        )

        sample[..., upp_bounded] = (
            -self.np_random.exponential(size=bdim + upp_bounded[upp_bounded].shape)
            - self.high[upp_bounded]
        )

        sample[..., bounded] = self.np_random.uniform(
            low=self.low[bounded],
            high=high[bounded],
            size=bdim + bounded[bounded].shape,
        )

        return sample.astype(self.dtype)

    def __repr__(self):
        # Prints the name of the class and its information
        # Specifically, the shape, the max of self.high, and the min of self.low
        # :return: string representation of the Box
        return "Box(low={0}, high={1}, shape={2})".format(
            np.min(self.low), np.max(self.high), self.shape
        )


class Discrete(gym.spaces.Discrete, MineRLSpace):
    def __init__(self, *args, **kwargs):
        super(Discrete, self).__init__(*args, **kwargs)
        self.eye = np.eye(self.n, dtype=np.float32)
        self._shape = ()

    def no_op(self, batch_shape=()):
        if len(batch_shape) == 0:
            return 0
        else:
            return (np.zeros(batch_shape)).astype(self.dtype)

    def create_flattened_space(self):
        return Box(low=0, high=1, shape=(self.n,))

    def flat_map(self, x):
        return self.eye[x]

    def unmap(self, x):
        return np.array(np.argmax(x, axis=-1), dtype=self.dtype)

    def sample(self, bs=None):
        bdim = () if bs is None else (bs,)
        return self.np_random.randint(self.n, size=bdim)


class Enum(Discrete, MineRLSpace):
    """
    An enum space. It can either be the enum string or an integer.
    """

    def __init__(self, *values: str, default=None):
        """Initializes the Enum space with a set of possible
        values that the enum can take.

        Usage:
        >>> x = Enum('none', 'type1', 'type2')
        >>> x['none']
        0
        >>> x['type1']
        1

        Args:
            values (str):  An order argument list of values the enum can take.
        """
        if not isinstance(values, tuple):
            values = (values,)
        self.default = default if default is not None else values[0]
        super().__init__(len(values))
        self.values = np.array(sorted(values))
        self.value_map = dict(zip(self.values, range(len(values))))

    def sample(self, bs=None) -> int:
        """Samples a random index for one of the enum types.

        ```
        x.sample() # A random nubmer in the half-open discrete interval [0, len(x.values))
        ````

        Returns:
            int:  A random index for one of the enum types.
        """
        return self.values[super().sample(bs)]

    def flat_map(self, x):
        return super().flat_map(self[x])

    def unmap(self, x):
        return self.values[super().unmap(x)]

    def no_op(self, batch_shape=()):
        if self.default:
            if len(batch_shape) == 0:
                return self.default
            else:
                return self.values[
                    super().no_op(batch_shape) + self.value_map[self.default]
                ]
        else:
            return self.values[super().no_op(batch_shape)]

    def __getitem__(self, action):
        try:
            single_act = False
            if isinstance(action, str):
                single_act = True
                action = np.array([action])

            u, inv = np.unique(action, return_inverse=True)

            inds = np.array([self.value_map[x] for x in u])[inv].reshape(action.shape)

            return inds if not single_act else inds.tolist()[0]

        except ValueError:
            raise ValueError(
                '"{}" not valid ENUM value in values {}'.format(action, self.values)
            )

        # TODO support more action formats through np.all < super().n
        raise ValueError("spaces.Enum: action must be of type str or int")

    def __str__(self):
        return "Enum(" + ",".join(self.values) + ")"

    def __repr__(self):
        return str(self)

    def __len__(self):
        return len(self.values)

    def contains(self, x):
        return x in self.values

    __contains__ = contains


# TODO: Vectorize containment?
class Dict(gym.spaces.Dict, MineRLSpace):
    def no_op(self, batch_shape=()):
        return OrderedDict(
            [
                (k, space.no_op(batch_shape=batch_shape))
                for k, space in self.spaces.items()
            ]
        )

    def create_flattened_space(self):
        shape = sum(
            [s.flattened.shape[0] for s in self.spaces.values() if s.is_flattenable()]
        )
        return Box(low=0, high=1, shape=[shape], dtype=np.float32)

    def create_unflattened_space(self):
        # TODO Fix this really ugly hack for flattening.
        # Needs to be a generic design that's simple that
        # encapsulates unflattenable or not;
        # First calss support for unflattenable spaces..
        return Dict(
            {
                k: (v.unflattened if hasattr(v, "unflattened") else v)
                for k, v in self.spaces.items()
                if not v.is_flattenable()
            }
        )

    def sample(self, bs=None):
        return OrderedDict([(k, v.sample(bs)) for k, v in self.spaces.items()])

    @property
    def unflattened(self):
        """
        Returns the unflatteneable part of the space.
        """
        if not hasattr(self, "_unflattened"):
            self._unflattened = self.create_unflattened_space()
        return self._unflattened

    def flat_map(self, x):
        # This could be refactored and externalized.
        # TODO: 1. Make all spaces have a shape.
        # TODO: 2. Make vectorization a first class citizen gym3?

        try:
            batch_shape = ()

            for (k, v) in self.spaces.items():
                if k in x and v.is_flattenable():
                    # Get batch_shape from x[k]
                    if not hasattr(x[k], "shape"):
                        # If any x[k] is a python prim well then clearly
                        # we are not vectorized; so there is no batch_size.
                        break
                    try:
                        batch_shape = (
                            x[k].shape
                            if len(v.shape) == 0
                            else x[k].shape[: -len(v.shape)]
                        )
                        break
                    except AttributeError:
                        # Cannot determine batch_size from vector without shape
                        pass

            stuff_to_cat = []
            for (k, v) in self.spaces.items():
                if v.is_flattenable():
                    stuff_to_cat.append(
                        (
                            v.flat_map(x[k])
                            if k in x
                            else v.flat_map(v.no_op(batch_shape))
                        )
                    )
            return np.concatenate(stuff_to_cat, axis=-1)
        except ValueError as e:
            # No flattenable handlers found
            return np.array([])

    def unflattenable_map(self, x: OrderedDict) -> OrderedDict:
        """
        Selects the unflattened part of x
        """
        return OrderedDict(
            {
                k: (
                    v.unflattenable_map(x[k])
                    if hasattr(v, "unflattenable_map")
                    else x[k]
                )
                # filter
                for k, v in (self.spaces.items())
                if not v.is_flattenable()
            }
        )

    def unmap(self, x: np.ndarray, skip=False) -> OrderedDict:
        unmapped = OrderedDict()
        cur_index = 0
        for k, v in self.spaces.items():
            if v.is_flattenable():
                unmapped[k] = v.unmap(
                    x[..., cur_index : cur_index + v.flattened.shape[0]]
                )
                cur_index += v.flattened.shape[0]
            elif not skip:
                raise ValueError(
                    "Dict space contains is_flattenable values - unmap with unmap_mixed"
                )

        return unmapped

    def unmap_mixed(self, x: np.ndarray, aux: OrderedDict):
        # split x
        unmapped = OrderedDict()
        cur_index = 0
        for k, v in self.spaces.items():
            if v.is_flattenable():
                try:
                    unmapped[k] = v.unmap_mixed(
                        x[..., cur_index : cur_index + v.flattened.shape[0]], aux[k]
                    )
                except (KeyError, AttributeError):
                    unmapped[k] = v.unmap(
                        x[..., cur_index : cur_index + v.flattened.shape[0]]
                    )
                cur_index += v.flattened.shape[0]
            else:
                unmapped[k] = aux[k]

        return unmapped

    def __delitem__(self, key):
        del self.spaces[key]


class MultiDiscrete(gym.spaces.MultiDiscrete, MineRLSpace):
    def __init__(self, *args, noop_vec=None, **kwargs):
        super(MultiDiscrete, self).__init__(*args, **kwargs)
        self.eyes = [np.eye(n, dtype=np.float32) for n in self.nvec]
        if noop_vec is not None:
            noop_vec = np.asarray(noop_vec, dtype=self.dtype)
            assert noop_vec.shape == self.nvec.shape
            assert self.contains(noop_vec)
        self._noop_vec = noop_vec

    def no_op(self, batch_shape=()):
        return (
            (np.zeros(list(batch_shape) + list(self.nvec.shape)) * self.nvec).astype(
                self.dtype
            )
            if self._noop_vec is None
            else (
                np.ones(list(batch_shape) + list(self.nvec.shape)) * self._noop_vec
            ).astype(self.dtype)
        )

    def create_flattened_space(self):
        return Box(low=0, high=1, shape=[np.sum(self.nvec)])

    def flat_map(self, x):
        return np.concatenate(
            [self.eyes[i][x[..., i]] for i in range(len(self.nvec))], axis=-1
        )

    def unmap(self, x):
        cur_index = 0
        out = []
        for n in self.nvec:
            out.append(
                np.argmax(x[..., cur_index : cur_index + n], axis=-1)[..., np.newaxis]
            )
            cur_index += n
        return np.concatenate(out, axis=-1).astype(self.dtype)

    def sample(self, bs=None):
        bdim = () if bs is None else (bs,)
        return (
            self.np_random.random_sample(bdim + self.nvec.shape) * self.nvec
        ).astype(self.dtype)


class Text(MineRLSpace):
    """
    # TODO:
    [['a text string', ..., 'last_text_string']]
    Example usage:
    self.observation_space = spaces.Text(1)
    """

    def no_op(self, batch_shape=()):
        return ""

    def create_flattened_space(self):
        raise NotImplementedError

    def flat_map(self, x):
        raise NotImplementedError

    def unmap(self, x):
        raise NotImplementedError

    MAX_STR_LEN = 100

    def __init__(self, shape):
        super().__init__(shape, np.unicode_)

    def sample(self, bs=None):
        total_strings = np.prod(self.shape)
        strings = [
            "".join(
                [
                    random.choice(string.ascii_lowercase)
                    for _ in range(random.randint(0, Text.MAX_STR_LEN))
                ]
            )
            for _ in range(total_strings)
        ]
        return np.array(np.reshape(strings, self.shape), np.dtype)

    def contains(self, x):
        contained = False  # ? TODO (R): Look back in git.
        contained = (
            contained
            or isinstance(x, np.ndarray)
            and x.shape == self.shape
            and x.dtype.type in [np.string_, np.unicode]
        )
        contained = contained or self.shape in [None, 1] and isinstance(x, str)
        return contained

    __contains__ = contains

    def to_jsonable(self, sample_n):
        return np.array(sample_n, dtype=self.dtype).to_list()

    def from_jsonable(self, sample_n):
        return [np.asarray(sample, dtype=self.dtype) for sample in sample_n]

    def __repr__(self):
        return "Text" + str(self.shape)

    def is_flattenable(self):
        return False


class DiscreteRange(Discrete):
    """
    {begin, begin+1, ..., end-2, end - 1}

    Like discrete, but takes a range of dudes
    DiscreteRange(0, n) is equivalent to Discrete(n)

    Examples usage:
    self.observation_space = spaces.DiscreteRange(-1, 3)
    """

    def __init__(self, begin, end):
        self.begin = begin
        self.end = end
        super().__init__(end - begin)

    def sample(self, bs=None):
        return super().sample(bs) + self.begin

    def contains(self, x):
        return super().contains(x - self.begin)

    __contains__ = contains

    def no_op(self, batch_shape=()):
        if len(batch_shape) == 0:
            return self.begin
        else:
            return (np.zeros(batch_shape) + self.begin).astype(self.dtype)

    def create_flattened_space(self):
        return Box(low=0, high=1, shape=(self.n,))

    def flat_map(self, x):
        return self.eye[x - self.begin]

    def unmap(self, x):
        return np.array(np.argmax(x, axis=-1) + self.begin, dtype=self.dtype)

    def __repr__(self):
        return "DiscreteRange({}, {})".format(self.begin, self.n + self.begin)

    def __eq__(self, other):
        return self.n == other.n and self.begin == other.begin



from functools import partial
from mypy_extensions import Arg
from typing import Union, Callable, Dict


__all__ = [
    "reward_fn_base",
    "simple_inventory_based_reward",
    "simple_stat_kill_entity_based_reward",
    "possess_item_reward",
    "survive_per_day_reward",
    "survive_n_days_reward",
    "use_any_item_reward",
]


# takes an initial info dict (t = 0), a pre info dict (t - 1), a current info dict (t), and elapsed time-steps,
# return a scalar reward value
reward_fn_base = Callable[
    [
        Arg(dict, "ini_info_dict"),
        Arg(dict, "pre_info_dict"),
        Arg(dict, "cur_info_dict"),
        Arg(int, "elapsed_timesteps"),
    ],
    float,
]


def _simple_stat_kill_entity_based_reward(
    name: str,
    weight: Union[int, float],
    ini_info_dict: dict,
    pre_info_dict: dict,
    cur_info_dict: dict,
    elapsed_timesteps: int,
):
    """
    A simple reward based on increment in `info["stat"]["kill_entity"][{name}]`.
    """
    return weight * (
        cur_info_dict["stat"]["kill_entity"][name]
        - pre_info_dict["stat"]["kill_entity"][name]
    )


def simple_stat_kill_entity_based_reward(
    name: str, weight: Union[int, float], **kwargs
) -> reward_fn_base:
    return partial(_simple_stat_kill_entity_based_reward, name=name, weight=weight)


def _simple_inventory_based_reward(
    name: str,
    weight: Union[int, float],
    ini_info_dict: dict,
    pre_info_dict: dict,
    cur_info_dict: dict,
    elapsed_timesteps: int,
):
    """
    A simple reward based on increment in `info["inventory"]`
    """
    return (
        sum(
            [
                inv_item["quantity"]
                for inv_item in cur_info_dict["inventory"]
                if inv_item["name"] == name
            ]
        )
        - sum(
            [
                inv_item["quantity"]
                for inv_item in pre_info_dict["inventory"]
                if inv_item["name"] == name
            ]
        )
    ) * weight


def simple_inventory_based_reward(
    name: str, weight: Union[int, float], **kwargs
) -> reward_fn_base:
    return partial(_simple_inventory_based_reward, name=name, weight=weight)


def _possess_item_reward(
    name: str,
    weight: Union[int, float],
    quantity: int,
    ini_info_dict: dict,
    pre_info_dict: dict,
    cur_info_dict: dict,
    elapsed_timesteps: int,
):
    return (
        float(
            (
                sum(
                    [
                        inv_item["quantity"]
                        for inv_item in cur_info_dict["inventory"]
                        if inv_item["name"] == name
                    ]
                )
                - sum(
                    [
                        inv_item["quantity"]
                        for inv_item in ini_info_dict["inventory"]
                        if inv_item["name"] == name
                    ]
                )
            )
            >= quantity
        )
        * weight
    )


def possess_item_reward(
    name: str, quantity: int, weight: Union[int, float]
) -> reward_fn_base:
    return partial(_possess_item_reward, name=name, quantity=quantity, weight=weight)


def _survive_per_day_reward(
    mc_ticks_per_day: int,
    weight: Union[int, float],
    ini_info_dict: dict,
    pre_info_dict: dict,
    cur_info_dict: dict,
    elapsed_timesteps: int,
):
    time_since_death_pre = pre_info_dict["stat"]["time_since_death"]
    time_since_death_cur = cur_info_dict["stat"]["time_since_death"]
    survived_days_pre = time_since_death_pre // mc_ticks_per_day
    survived_days_cur = time_since_death_cur // mc_ticks_per_day
    return (survived_days_cur - survived_days_pre) * weight


def survive_per_day_reward(
    mc_ticks_per_day: int, weight: Union[int, float]
) -> reward_fn_base:
    return partial(
        _survive_per_day_reward, mc_ticks_per_day=mc_ticks_per_day, weight=weight
    )


def _survive_n_days_reward(
    mc_ticks_per_day: int,
    target_days: int,
    weight: Union[int, float],
    ini_info_dict: dict,
    pre_info_dict: dict,
    cur_info_dict: dict,
    elapsed_timesteps: int,
):
    return weight * float(
        cur_info_dict["stat"]["time_since_death"] >= mc_ticks_per_day * target_days
    )


def survive_n_days_reward(
    mc_ticks_per_day: int, target_days: int, weight: Union[int, float]
) -> reward_fn_base:
    return partial(
        _survive_n_days_reward,
        mc_ticks_per_day=mc_ticks_per_day,
        target_days=target_days,
        weight=weight,
    )


def _use_any_item_reward(
    items_and_weights: Dict[str, Union[int, float]],
    ini_info_dict: dict,
    pre_info_dict: dict,
    cur_info_dict: dict,
    elapsed_timesteps: int,
):
    """
    reward any usage of item in items_and_weights.keys()
    """
    return sum(
        [
            (
                cur_info_dict["stat"]["use_item"]["minecraft"][item]
                - pre_info_dict["stat"]["use_item"]["minecraft"][item]
            )
            * weight
            for item, weight in items_and_weights.items()
        ]
    )


def use_any_item_reward(
    items_and_weights: Dict[str, Union[int, float]]
) -> reward_fn_base:
    return partial(_use_any_item_reward, items_and_weights=items_and_weights)



# Copyright (c) 2020 All Rights Reserved
# Author: William H. Guss, Brandon Houghton

from .action import *
from .actions import *
from .observations import *

from .quit import *
from .start import *



from .harvest import HarvestMeta
from .playthrough import Playthrough
from .combat import CombatMeta
from .survival import SurvivalMeta
from .tech_tree import TechTreeMeta
from .creative import CreativeMeta



"""
Do fast reset: reset using MC native command `/kill`, instead of waiting for generating new worlds

Side effects:
    1. Changes to the world will not be reset. E.g., if the agent chops lots of trees then calling fast reset
        will not restore those trees.
    2. If you specify agent starting health and food, these specs will only be respected at the first reset
        (i.e., generating a new world) but will not be respected in the following resets (i.e., reset using MC cmds).
        So be careful to use this wrapper if your usages require specific initial health/food.
    3. Statistics/achievements will not be reset. This wrapper will maintain a property `info_prev_reset`. If your
        tasks use stat/achievements to evaluation, please retrieve this property and calculate differences
"""
from __future__ import annotations

import gym

from ..sim import MineDojoSim
from ...sim.mc_meta.mc import MAX_FOOD, MAX_LIFE
from ..inventory import parse_inventory_item, map_slot_number_to_cmd_slot


class FastResetWrapper(gym.Wrapper):
    def __init__(
        self,
        env: MineDojoSim,
        random_teleport_range: int | None = 200,
        clear_ground: bool = True,
    ):
        super().__init__(env=env)
        start_time, start_weather = env.start_time, env.initial_weather
        initial_inventory, start_position = env.initial_inventory, env.start_position
        start_health, start_food = env.start_health, env.start_food
        if start_health != MAX_LIFE or start_food != MAX_FOOD:
            print(
                "Warning! You use non-default values for `start_health` and `start_food`. "
                "However, they will not take effects because `fast_reset = True`. "
                "Consider using `fast_reset = False` instead."
            )
        if random_teleport_range is None:
            random_teleport_range = 200
        assert random_teleport_range >= 0
        self._reset_cmds = [
            "/kill",
        ]
        if random_teleport_range > 0:
            self._reset_cmds.append(
                f"/spreadplayers ~ ~ 0 {random_teleport_range} false @p"
            )
        self._reset_cmds.extend(
            [f"/time set {start_time or 0}", f'/weather {start_weather or "normal"}']
        )
        if initial_inventory is not None:
            for inventory_item in initial_inventory:
                slot, item_dict = parse_inventory_item(inventory_item)
                self._reset_cmds.append(
                    f'/replaceitem entity @p {map_slot_number_to_cmd_slot(slot)} minecraft:{item_dict["type"]} {item_dict["quantity"]} {item_dict["metadata"]}'
                )
        if start_position is not None:
            self._reset_cmds.append(
                f'/tp {start_position["x"]} {start_position["y"]} {start_position["z"]} {start_position["yaw"]} {start_position["pitch"]}'
            )
        if clear_ground:
            self._reset_cmds.append("/kill @e[type=item]")

        self._server_start = False
        self._info_prev_reset = None

    def reset(self):
        if not self._server_start:
            self._server_start = True
            return self.env.reset()
        else:
            for cmd in self._reset_cmds:
                obs, _, _, info = self.env.execute_cmd(cmd)
            self._info_prev_reset = self.env.prev_info
            return obs

    def execute_cmd(self, *args, **kwargs):
        return self.env.execute_cmd(*args, **kwargs)

    def spawn_mobs(self, *args, **kwargs):
        return self.env.spawn_mobs(*args, **kwargs)

    def set_block(self, *args, **kwargs):
        return self.env.set_block(*args, **kwargs)

    def clear_inventory(self, *args, **kwargs):
        return self.env.clear_inventory(*args, **kwargs)

    def set_inventory(self, *args, **kwargs):
        return self.env.set_inventory(*args, **kwargs)

    def teleport_agent(self, *args, **kwargs):
        return self.env.teleport_agent(*args, **kwargs)

    def kill_agent(self, *args, **kwargs):
        return self.env.kill_agent(*args, **kwargs)

    def set_time(self, *args, **kwargs):
        return self.env.set_time(*args, **kwargs)

    def set_weather(self, *args, **kwargs):
        return self.env.set_weather(*args, **kwargs)

    def random_teleport(self, *args, **kwargs):
        return self.env.random_teleport(*args, **kwargs)

    @property
    def prev_obs(self):
        return self.env.prev_obs

    @property
    def prev_info(self):
        return self.env.prev_info

    @property
    def info_prev_reset(self):
        return self._info_prev_reset

    @property
    def prev_action(self):
        return self.env.prev_action

    @property
    def is_terminated(self):
        return self.env.is_terminated



# Copyright (c) 2021 All Rights Reserved
# Author: Brandon Houghton
from typing import List

import numpy as np

from minedojo.sim import spaces
from minedojo.sim.handlers.translation import (
    KeymapTranslationHandler,
    TranslationHandlerGroup,
)


__all__ = ["ObservationFromDamageSource"]


class ObservationFromDamageSource(TranslationHandlerGroup):
    """
    Includes the most recent damage event including the amount, type, location, and
    other properties.
    """

    def xml_template(self) -> str:
        return ""

    def to_string(self) -> str:
        return "damage_source"

    def __init__(self):
        super().__init__(
            handlers=[
                _DamageAmount(),
                _DamageDistance(),
                _DamagePitch(),
                _DamageYaw(),
                _HungerDamage(),
                _IsExplosion(),
                _IsFire(),
                _IsMagic(),
                _IsProjectile(),
                _IsUnblockable(),
            ]
        )


class _DamageSourceProperty(KeymapTranslationHandler):
    def to_hero(self, x) -> int:
        for key in self.hero_keys:
            x = x[key]
        return x

    def __init__(self, key_list: List[str], space=None, default_if_missing=None):
        if space is None:
            space = spaces.Box(low=0, high=np.inf, shape=(), dtype=np.int)
        if default_if_missing is None:
            default_if_missing = np.zeros((), dtype=space.dtype)

        super().__init__(
            hero_keys=key_list,
            univ_keys=key_list,
            space=space,
            default_if_missing=default_if_missing,
            ignore_missing=True,
        )

    def xml_template(self) -> str:
        return ""


class _DamageAmount(_DamageSourceProperty):
    def __init__(self):
        super().__init__(
            key_list=["damage_amount"],
            space=spaces.Box(low=0.0, high=40.0, shape=(1,), dtype=np.float32),
        )

    def from_hero(self, hero_dict, dtype=None):
        return np.expand_dims(super().from_hero(hero_dict, dtype), axis=0).astype(
            self.space.dtype
        )


class _DamageDistance(_DamageSourceProperty):
    def __init__(self):
        super().__init__(
            key_list=["damage_distance"],
            space=spaces.Box(low=0.0, high=np.inf, shape=(1,), dtype=np.float32),
        )

    def from_hero(self, hero_dict, dtype=None):
        return np.expand_dims(super().from_hero(hero_dict, dtype), axis=0).astype(
            self.space.dtype
        )


class _HungerDamage(_DamageSourceProperty):
    def __init__(self):
        super().__init__(
            key_list=["hunger_damage"],
            space=spaces.Box(low=0, high=20, shape=(1,), dtype=np.float32),
        )

    def from_hero(self, hero_dict, dtype=None):
        return np.expand_dims(super().from_hero(hero_dict, dtype), axis=0).astype(
            self.space.dtype
        )


class _DamagePitch(_DamageSourceProperty):
    def __init__(self):
        super().__init__(
            key_list=["damage_pitch"],
            space=spaces.Box(low=-90, high=90, shape=(1,), dtype=np.float32),
        )

    def from_hero(self, hero_dict, dtype=None):
        return np.expand_dims(super().from_hero(hero_dict, dtype), axis=0).astype(
            self.space.dtype
        )


class _DamageYaw(_DamageSourceProperty):
    def __init__(self):
        super().__init__(
            key_list=["damage_yaw"],
            space=spaces.Box(low=-180, high=180, shape=(1,), dtype=np.float32),
        )

    def from_hero(self, hero_dict, dtype=None):
        return np.expand_dims(super().from_hero(hero_dict, dtype), axis=0).astype(
            self.space.dtype
        )


class _IsExplosion(_DamageSourceProperty):
    def __init__(self):
        super().__init__(
            key_list=["is_explosion"],
            space=spaces.Box(low=0, high=1, shape=(), dtype=bool),
        )


class _IsFire(_DamageSourceProperty):
    def __init__(self):
        super().__init__(
            key_list=["is_fire_damage"],
            space=spaces.Box(low=0, high=1, shape=(), dtype=bool),
        )


class _IsMagic(_DamageSourceProperty):
    def __init__(self):
        super().__init__(
            key_list=["is_magic_damage"],
            space=spaces.Box(low=0, high=1, shape=(), dtype=bool),
        )


class _IsProjectile(_DamageSourceProperty):
    def __init__(self):
        super().__init__(
            key_list=["is_projectile"],
            space=spaces.Box(low=0, high=1, shape=(), dtype=bool),
        )


class _IsUnblockable(_DamageSourceProperty):
    def __init__(self):
        super().__init__(
            key_list=["is_unblockable"],
            space=spaces.Box(low=0, high=1, shape=(), dtype=bool),
        )



from functools import partial
from mypy_extensions import Arg
from typing import Union, Callable


__all__ = [
    "extra_spawn_condition_base",
    "always_satisfy_condition",
    "check_below_height_condition",
    "check_above_height_condition",
    "check_below_light_level_condition",
    "check_above_light_level_condition",
]


# takes an initial info dict (t = 0), a pre info dict (t - 1), and elapsed time-steps,
# return if the condition is satisfied at the current step (t)
extra_spawn_condition_base = Callable[
    [
        Arg(dict, "ini_info_dict"),
        Arg(dict, "pre_info_dict"),
        Arg(int, "elapsed_timesteps"),
    ],
    bool,
]


def always_satisfy_condition(
    ini_info_dict: dict, pre_info_dict: dict, elapsed_timesteps: int
) -> bool:
    return True


def _check_below_height_condition(
    height_threshold: Union[float, int],
    ini_info_dict: dict,
    pre_info_dict: dict,
    elapsed_timesteps: int,
):
    """
    Satisfy condition iff info["ypos"] <= height_threshold.
    Note that in MC, y-axis is the vertical axis.
    This is useful for extra diamond ore spawn (below y = 14)
    """
    return pre_info_dict["ypos"] <= height_threshold


def check_below_height_condition(
    height_threshold: Union[float, int]
) -> extra_spawn_condition_base:
    return partial(_check_below_height_condition, height_threshold=height_threshold)


def _check_above_height_condition(
    height_threshold: Union[float, int],
    ini_info_dict: dict,
    pre_info_dict: dict,
    elapsed_timesteps: int,
):
    """
    Satisfy condition iff info["ypos"] >= height_threshold.
    Note that in MC, y-axis is the vertical axis.
    This is useful for extra plants spawn (above sea level y >= 62)
    """
    return pre_info_dict["ypos"] >= height_threshold


def check_above_height_condition(
    height_threshold: Union[float, int]
) -> extra_spawn_condition_base:
    return partial(_check_above_height_condition, height_threshold=height_threshold)


def _check_below_light_level_condition(
    light_level_threshold: Union[float, int],
    ini_info_dict: dict,
    pre_info_dict: dict,
    elapsed_timesteps: int,
):
    """
    Satisfy condition iff info["light_level"] <= light_level_threshold
    Userful for extra monsters spawn (only spawn in dark/night)
    """
    return pre_info_dict["light_level"] <= light_level_threshold


def check_below_light_level_condition(light_level_threshold: Union[float, int]):
    return partial(
        _check_below_light_level_condition, light_level_threshold=light_level_threshold
    )


def _check_above_light_level_condition(
    light_level_threshold: Union[float, int],
    ini_info_dict: dict,
    pre_info_dict: dict,
    elapsed_timesteps: int,
):
    """
    Satisfy condition iff info["light_level"] >= light_level_threshold
    Userful for extra passive mobs spawn
    """
    return pre_info_dict["light_level"] >= light_level_threshold


def check_above_light_level_condition(light_level_threshold: Union[float, int]):
    return partial(
        _check_above_light_level_condition, light_level_threshold=light_level_threshold
    )



"""These handlers allow episode termination based on server conditions (e.g. time passed)

When used to create a Gym environment, they should be passed to :code:`create_server_quit_producers`
"""
# Copyright (c) 2020 All Rights Reserved
# Author: William H. Guss, Brandon Houghton


from minedojo.sim.handler import Handler


class ServerQuitFromTimeUp(Handler):
    """
    Forces the server to quit after a certain time_limit_ms
    also specifies a description parameter for the xml.

    Example usage

    .. code-block:: python

        ServerQuitFromTimeUp(50000)
    """

    def to_string(self) -> str:
        return "server_quit_after_time_up"

    def xml_template(self) -> str:
        return str(
            """<ServerQuitFromTimeUp 
                    timeLimitMs="{{ time_limit_ms | string }}"
                    description="{{description}}"/>
            """
        )

    def __init__(self, time_limit_ms: int, description="out_of_time"):
        self.time_limit_ms = time_limit_ms
        self.description = description


class ServerQuitFromDeath(Handler):
    """Forces the server to quit after a certain time_limit_ms
    also specifies a description parameter for the xml."""

    def to_string(self) -> str:
        return "server_quit_from_death"

    def xml_template(self) -> str:
        return str(
            """<ServerQuitFromTimeUp 
                    quitWhenAnyDead="{{quit_when_any_agent_dies}}"/>
            """
        )

    def __init__(self, quit_when_any_agent_dies: bool):
        self.quit_when_any_agent_dies = quit_when_any_agent_dies


class ServerQuitWhenAnyAgentFinishes(Handler):
    """
    Forces the server to quit if any of the agents involved quits.
    Has no parameters.

    Example usage:

    .. code-block:: python

        ServerQuitWhenAnyAgentFinishes()
    """

    def to_string(self) -> str:
        return "server_quit_when_any_agent_finishes"

    def xml_template(self) -> str:
        return str(
            """<ServerQuitWhenAnyAgentFinishes/>
            """
        )



# Copyright (c) 2020 All Rights Reserved
# Author: William H. Guss, Brandon Houghton


from .camera import *
from .craft import *
from .equip import *
from .keyboard import *
from .place import *
from .smelt import *
from .swap_slot import SwapSlotAction



import math
from typing import Union, Sequence

import gym
import numpy as np

from ...sim import MineDojoSim
from ....sim import spaces as spaces
from ....sim.mc_meta import mc as MC
from ....sim.inventory import InventoryItem


class NNActionSpaceWrapper(gym.Wrapper):
    """
    Action wrapper to transform native action space to a new space friendly to train NNs
    """

    def __init__(
        self,
        env: Union[MineDojoSim, gym.Wrapper],
        discretized_camera_interval: Union[int, float] = 15,
        strict_check: bool = True,
    ):
        assert (
            "equip" in env.action_space.keys()
            and "place" in env.action_space.keys()
            and "swap_slot" not in env.action_space.keys()
        ), "please use this wrapper with event_level_control = True"
        assert (
            "inventory" in env.observation_space.keys()
        ), f"missing inventory from obs space"
        super().__init__(env=env)

        n_pitch_bins = math.ceil(360 / discretized_camera_interval) + 1
        n_yaw_bins = math.ceil(360 / discretized_camera_interval) + 1

        self.action_space = spaces.MultiDiscrete(
            [
                3,  # forward and back, 0: noop, 1: forward, 2: back
                3,  # 0: noop, 1: left, 2: right
                4,  # 0: noop, 1: jump, 2: sneak, 3: sprint
                n_pitch_bins,  # camera pitch, 0: -180, n_pitch_bins - 1: +180
                n_yaw_bins,  # camera yaw, 0: -180, n_yaw_bins - 1: +180,
                8,  # functional actions, 0: no_op, 1: use, 2: drop, 3: attack 4: craft 5: equip 6: place 7: destroy
                len(MC.ALL_CRAFT_SMELT_ITEMS),  # arg for "craft"
                MC.N_INV_SLOTS,  # arg for "equip", "place", and "destroy"
            ],
            noop_vec=[
                0,
                0,
                0,
                (n_pitch_bins - 1) // 2,
                (n_yaw_bins - 1) // 2,
                0,
                0,
                0,
            ],
        )
        self._cam_interval = discretized_camera_interval
        self._inventory_names = None
        self._strict_check = strict_check

    def action(self, action: Sequence[int]):
        """
        NN action to Malmo action
        """
        assert self.action_space.contains(action)
        destroy_item = (False, None)
        noop = self.env.action_space.no_op()

        # ------ parse main actions ------
        # parse forward and back
        if action[0] == 1:
            noop["forward"] = 1
        elif action[0] == 2:
            noop["back"] = 1
        # parse left and right
        if action[1] == 1:
            noop["left"] = 1
        elif action[1] == 2:
            noop["right"] = 1
        # parse jump sneak and sprint
        if action[2] == 1:
            noop["jump"] = 1
        elif action[2] == 2:
            noop["sneak"] = 1
        elif action[2] == 3:
            noop["sprint"] = 1
        # parse camera pitch
        noop["camera"][0] = float(action[3]) * self._cam_interval + (-180)
        # parse camera yaw
        noop["camera"][1] = float(action[4]) * self._cam_interval + (-180)

        # ------ parse functional actions ------
        fn_action = action[5]
        # note that 0 is no_op
        if fn_action == 0:
            pass
        elif fn_action == 1:
            noop["use"] = 1
        elif fn_action == 2:
            noop["drop"] = 1
        elif fn_action == 3:
            noop["attack"] = 1
        elif fn_action == 4:
            item_to_craft = MC.ALL_CRAFT_SMELT_ITEMS[action[6]]
            if item_to_craft in MC.ALL_HAND_CRAFT_ITEMS_NN_ACTIONS:
                noop["craft"] = item_to_craft
            elif item_to_craft in MC.ALL_TABLE_CRAFT_ONLY_ITEMS_NN_ACTIONS:
                noop["craft_with_table"] = item_to_craft
            elif item_to_craft in MC.ALL_SMELT_ITEMS_NN_ACTIONS:
                noop["smelt"] = item_to_craft
            elif self._strict_check:
                raise ValueError(f"Unknown item {item_to_craft} to craft/smelt!")
        elif fn_action == 5:
            assert action[7] in list(range(MC.N_INV_SLOTS))
            item_id = self._inventory_names[action[7]].replace(" ", "_")
            if item_id == "air":
                if self._strict_check:
                    raise ValueError(
                        "Trying to equip air, raise error with strict check."
                        "You shouldn't execute this action, maybe something wrong with the mask!"
                    )
            else:
                noop["equip"] = item_id
        elif fn_action == 6:
            assert action[7] in list(range(MC.N_INV_SLOTS))
            item_id = self._inventory_names[action[7]].replace(" ", "_")
            if item_id == "air":
                if self._strict_check:
                    raise ValueError(
                        "Trying to place air, raise error with strict check."
                        "You shouldn't execute this action, maybe something wrong with the mask!"
                    )
            else:
                noop["place"] = item_id
        elif fn_action == 7:
            assert action[7] in list(range(MC.N_INV_SLOTS))
            item_id = self._inventory_names[action[7]].replace(" ", "_")
            if item_id == "air":
                if self._strict_check:
                    raise ValueError(
                        "Trying to destroy air, raise error with strict check."
                        "You shouldn't execute this action, maybe something wrong with the mask!"
                    )
            else:
                destroy_item = (True, action[7])
        else:
            raise ValueError(f"Unknown value {fn_action} for function action")
        return noop, destroy_item

    def reverse_action(self, action):
        """
        Malmo action to NN action
        """
        # first convert camera actions to [-pi, +pi]
        action["camera"] = (
            np.arctan2(
                np.sin(action["camera"] * np.pi / 180),
                np.cos(action["camera"] * np.pi / 180),
            )
            * 180
            / np.pi
        )
        assert self.env.action_space.contains(action)

        noop = self.action_space.no_op()
        # ------ parse main actions ------
        # parse forward and back
        if action["forward"] == 1 and action["back"] == 1:
            # cancel each other, noop
            pass
        elif action["forward"] == 1:
            noop[0] = 1
        elif action["back"] == 1:
            noop[0] = 2
        # parse left and right
        if action["left"] == 1 and action["right"] == 1:
            # cancel each other, noop
            pass
        elif action["left"] == 1:
            noop[1] = 1
        elif action["right"] == 1:
            noop[1] = 2
        # parse jump, sneak, sprint
        # prioritize jump
        if action["jump"] == 1:
            noop[2] = 1
        else:
            if action["sneak"] == 1 and action["sprint"] == 1:
                # cancel each other, noop
                pass
            elif action["sneak"] == 1:
                noop[2] = 2
            elif action["sprint"] == 1:
                noop[2] = 3
        # parse camera pitch
        noop[3] = math.ceil((action["camera"][0] - (-180)) / self._cam_interval)
        # parse camera yaw
        noop[4] = math.ceil((action["camera"][1] - (-180)) / self._cam_interval)

        # ------ parse functional actions ------
        # order: attack > use > craft > equip > place > drop > destroy
        if action["attack"] == 1:
            noop[5] = 3
        elif action["use"] == 1:
            noop[5] = 1
        elif action["craft"] != "none" and action["craft"] != 0:
            craft = action["craft"]
            if isinstance(craft, int):
                craft = MC.ALL_PERSONAL_CRAFTING_ITEMS[craft - 1]
            noop[5] = 4
            noop[6] = MC.ALL_CRAFT_SMELT_ITEMS.index(craft)
        elif action["craft_with_table"] != "none" and action["craft_with_table"] != 0:
            craft = action["craft_with_table"]
            if isinstance(craft, int):
                craft = MC.ALL_CRAFTING_TABLE_ITEMS[craft - 1]
            noop[5] = 4
            noop[6] = MC.ALL_CRAFT_SMELT_ITEMS.index(craft)
        elif action["smelt"] != "none" and action["smelt"] != 0:
            smelt = action["smelt"]
            if isinstance(smelt, int):
                smelt = MC.ALL_SMELTING_ITEMS[smelt - 1]
            noop[5] = 4
            noop[6] = MC.ALL_CRAFT_SMELT_ITEMS.index(smelt)
        elif action["equip"] != "none" and action["equip"] != 0:
            equip = action["equip"]
            if isinstance(equip, int):
                equip = MC.ALL_ITEMS[equip - 1]
            equip = equip.replace("_", " ")
            if equip not in self._inventory_names:
                if self._strict_check:
                    raise ValueError(
                        f"try to equip {equip}, but it is not in the inventory {self._inventory_names}"
                    )
            else:
                slot_idx = np.where(self._inventory_names == equip)[0][0]
                noop[5] = 5
                noop[7] = slot_idx
        elif action["place"] != "none" and action["place"] != 0:
            place = action["place"]
            if isinstance(place, int):
                place = MC.ALL_ITEMS[place - 1]
            place = place.replace("_", " ")
            if place not in self._inventory_names:
                if self._strict_check:
                    raise ValueError(
                        f"try to place {place}, but it is not in the inventory {self._inventory_names}"
                    )
            else:
                slot_idx = np.where(self._inventory_names == place)[0][0]
                noop[5] = 6
                noop[7] = slot_idx
        elif action["drop"] == 1:
            noop[5] = 2
        return noop

    def reset(self, **kwargs):
        obs = self.env.reset(**kwargs)
        self._inventory_names = obs["inventory"]["name"].copy()
        return obs

    def step(self, action: Sequence[int]):
        malmo_action, destroy_item = self.action(action)
        destroy_item, destroy_slot = destroy_item
        if destroy_item:
            obs, reward, done, info = self.env.set_inventory(
                inventory_list=[
                    InventoryItem(name="air", slot=destroy_slot, quantity=1, variant=0)
                ],
                action=malmo_action,
            )
        else:
            obs, reward, done, info = self.env.step(malmo_action)

        # handle malmo's lags
        if action[5] in {2, 4, 5, 6, 7}:
            for _ in range(2):
                obs, reward, done, info = self.env.step(self.env.action_space.no_op())
        self._inventory_names = obs["inventory"]["name"].copy()
        return obs, reward, done, info



from typing import Any, Union, Optional

import gym
import numpy as np

from ...sim import MineDojoSim
from ....sim import spaces as spaces
from ....sim.mc_meta import mc as MC
from ..utils import get_recipes_matrix, get_inventory_vector


class ARMasksWrapper(gym.ObservationWrapper):
    def __init__(
        self,
        env: Union[MineDojoSim, gym.Wrapper],
        action_categories_and_num_args: Optional[dict[str, int]] = None,
    ):
        assert "inventory" in env.observation_space.keys()
        assert "nearby_tools" in env.observation_space.keys()
        assert "table" in env.observation_space["nearby_tools"].keys()
        assert "furnace" in env.observation_space["nearby_tools"].keys()
        assert isinstance(
            env.action_space, spaces.MultiDiscrete
        ), "please use this wrapper with `NNActionSpaceWrapper!`"
        assert (
            len(env.action_space.nvec) == 8
        ), "please use this wrapper with `NNActionSpaceWrapper!`"
        super().__init__(env=env)

        action_categories_and_num_args = action_categories_and_num_args or dict(
            no_op=0, use=0, drop=0, attack=0, craft=1, equip=1, place=1, destroy=1
        )
        obs_space = env.observation_space
        n_fn_actions = len(action_categories_and_num_args)
        obs_space["masks"] = spaces.Dict(
            {
                "action_type": spaces.Box(
                    low=0, high=1, shape=(n_fn_actions,), dtype=bool
                ),
                "action_arg": spaces.Box(
                    low=0, high=1, shape=(n_fn_actions, 1), dtype=bool
                ),
                "equip": spaces.Box(low=0, high=1, shape=(MC.N_INV_SLOTS,), dtype=bool),
                "place": spaces.Box(low=0, high=1, shape=(MC.N_INV_SLOTS,), dtype=bool),
                "destroy": spaces.Box(
                    low=0, high=1, shape=(MC.N_INV_SLOTS,), dtype=bool
                ),
                "craft_smelt": spaces.Box(
                    low=0, high=1, shape=(len(MC.ALL_CRAFT_SMELT_ITEMS),), dtype=bool
                ),
            }
        )
        # remove `full_stats`
        if "full_stats" in obs_space.keys():
            del obs_space["full_stats"]
        self.observation_space = obs_space
        # a_arg_mask can be determined now and will not change
        self._a_arg_mask = np.array(
            list(action_categories_and_num_args.values()), dtype=bool
        )[..., np.newaxis]
        self._a_cats = list(action_categories_and_num_args.keys())
        self._n_fn_actions = n_fn_actions
        # no_op, use, and attack are always valid
        self._a_cat_always_true_indices = np.isin(
            np.array(self._a_cats), np.array(["no_op", "use", "attack"])
        ).nonzero()[0]
        self._equip_place_destroy_indices = np.isin(
            np.array(self._a_cats), np.array(["equip", "place", "destroy"])
        ).nonzero()[0]
        # get recipe matrix
        self._recipes = get_recipes_matrix()

    def observation(self, observation: dict[str, Any]):
        # ------ craft smelt mask ------
        inventory_vector = get_inventory_vector(observation["inventory"])
        craft_smelt_mask = ~np.any(
            (inventory_vector - self._recipes) < 0, axis=1
        ).astype(
            bool
        )  # (N_all_craftable)
        # meta mask using `nearby_table` and `nearby_furnace`
        nearby_table, nearby_furnace = (
            observation["nearby_tools"]["table"],
            observation["nearby_tools"]["furnace"],
        )
        nearby_table = np.array(
            [nearby_table] * len(MC.ALL_TABLE_CRAFT_ONLY_ITEMS_NN_ACTIONS), dtype=bool
        )
        nearby_furnace = np.array(
            [nearby_furnace] * len(MC.ALL_SMELT_ITEMS_NN_ACTIONS), dtype=bool
        )
        table_craft_items_indices = slice(
            len(MC.ALL_HAND_CRAFT_ITEMS_NN_ACTIONS),
            len(MC.ALL_HAND_CRAFT_ITEMS_NN_ACTIONS)
            + len(MC.ALL_TABLE_CRAFT_ONLY_ITEMS_NN_ACTIONS),
        )
        craft_smelt_mask[table_craft_items_indices] = np.logical_and(
            nearby_table, craft_smelt_mask[table_craft_items_indices]
        )
        craft_smelt_mask[-len(MC.ALL_SMELT_ITEMS_NN_ACTIONS) :] = np.logical_and(
            nearby_furnace, craft_smelt_mask[-len(MC.ALL_SMELT_ITEMS_NN_ACTIONS) :]
        )
        # ------ determine destroy mask ------
        # destroy mask is simply if slots are occupied
        destroy_mask = (observation["inventory"]["name"] != "air").astype(bool)
        # ------ determine place mask ------
        # True if that slot is occupied by placeable items
        place_mask = np.logical_and(
            (observation["inventory"]["name"] != "air").astype(bool),
            np.isin(
                observation["inventory"]["name"], np.array(MC.PLACEABLE_ITEM_NAMES)
            ),
        )
        # ------ determine equip mask
        # True if that slot is occupied
        # special treat for main-hand slot
        equip_mask = (observation["inventory"]["name"] != "air").astype(bool)
        equip_mask[0] = (
            equip_mask[0]
            and observation["inventory"]["name"][0] in MC.NON_MAINHAND_ITEM_NAMES
        )

        # ------ determine action category mask ------
        a_cat_mask = np.empty(shape=(self._n_fn_actions,), dtype=bool)
        # no_op, use, and attack are always valid
        a_cat_mask[self._a_cat_always_true_indices] = True
        # validity of drop is dependent on if anything held in the main-hand
        a_cat_mask[self._a_cats.index("drop")] = (
            False if observation["inventory"]["name"][0] == "air" else True
        )
        # validity of craft equals to any(craft_smelt_mask) and the inventory is not full
        a_cat_mask[self._a_cats.index("craft")] = np.any(
            craft_smelt_mask
        ) and not np.all((observation["inventory"]["name"] != "air").astype(bool))
        # validity of equip simply equals to any(equip_mask)
        a_cat_mask[self._a_cats.index("equip")] = np.any(equip_mask)
        # validity of place simply equals to any(place_mask)
        a_cat_mask[self._a_cats.index("place")] = np.any(place_mask)
        # validity of destroy simple equals to any(destroy_mask)
        a_cat_mask[self._a_cats.index("destroy")] = np.any(destroy_mask)

        observation["masks"] = {
            "action_type": a_cat_mask,
            "action_arg": self._a_arg_mask,
            "equip": equip_mask,
            "place": place_mask,
            "destroy": destroy_mask,
            "craft_smelt": craft_smelt_mask,
        }
        # remove `full_stats`
        if "full_stats" in observation:
            del observation["full_stats"]
        return observation



# Copyright (c) 2020 All Rights Reserved
# Author: William H. Guss, Brandon Houghton

"""
Agent start handlers define agent start conditions such as inventory items and health.

When used to create a Gym environment, they should be passed to :code:`create_agent_start`
"""
import random
from typing import Dict, List, Union

from minedojo.sim.handler import Handler


class InventoryAgentStart(Handler):
    """
    Sets the start inventory of the agent by slot id.

    Example usage:

    .. code-block:: python

        InventoryAgentStart({
            0: {'type':'dirt', 'quantity':10},
            # metadata specifies the type of planks (e.g. oak, spruce)
            1: {'type':'planks', 'metadata': 1, 'quantity':5},
            5: {'type':'log', 'quantity':1},
            6: {'type':'log', 'quantity':2},
            32: {'type':'iron_ore', 'quantity':4
        })
    """

    def to_string(self) -> str:
        return "inventory_agent_start"

    def xml_template(self) -> str:
        return str(
            """<Inventory>
            {% for  slot in inventory %}
                <InventoryObject
                    slot="{{ slot }}"
                    type="{{ inventory[slot]['type'] }}"
                    metadata="{{ inventory[slot].get('metadata', 0) }}"
                    quantity="{{ inventory[slot]['quantity'] }}"
                    />
            {% endfor %}
            </Inventory>
            """
        )

    def __init__(self, inventory: Dict[int, Dict[str, Union[str, int]]]):
        """
        Args:
            inventory (Dict[int, Dict[str, Union[str,int]]]): The inventory slot description.
        """
        self.inventory = inventory


class SimpleInventoryAgentStart(InventoryAgentStart):
    """
    Sets the start inventory of the agent sequentially.

    Example usage:

    .. code-block:: python

        SimpleInventoryAgentStart([
            {'type':'dirt', 'quantity':10},
            {'type':'planks', 'quantity':5},
            {'type':'log', 'quantity':1},
            {'type':'iron_ore', 'quantity':4}
        ])
    """

    def __init__(self, inventory: List[Dict[str, Union[str, int]]]):
        """Creates a simple inventory agent start."""
        super().__init__({i: item for i, item in enumerate(inventory)})


class RandomInventoryAgentStart(InventoryAgentStart):
    """
    Sets the agent start inventory by randomly distributing items throughout its inventory slots.
    Note: This has no effect on inventory observation handlers.

    Example usage:

    .. code-block:: python

        RandomInventoryAgentStart(
            {'dirt': 10, 'planks': 5}
        )
    """

    def __init__(self, inventory: Dict[str, Union[str, int]], use_hotbar: bool = False):
        """Creates an inventory where items are placed in random positions"""
        self.inventory = inventory
        self.slot_range = (0, 36) if use_hotbar else (10, 36)

    def xml_template(self) -> str:
        lines = ["<Inventory>"]
        for item, quantity in self.inventory.items():
            slot = random.randint(*self.slot_range)
            lines.append(
                f'<InventoryObject slot="{slot}" type="{item}" quantity="{quantity}"/>'
            )
        lines.append("</Inventory>")
        return "\n".join(lines)


class AgentStartBreakSpeedMultiplier(Handler):
    """
    Sets the break speed multiplier (how fast the agent can break blocks)

    See here for more information: https://minecraft.fandom.com/el/wiki/Breaking

    Example usage:

    .. code-block:: python

        AgentStartBreakSpeedMultiplier(2.0)
    """

    def to_string(self) -> str:
        return f"agent_start_break_speed_multiplier({self.multiplier})"

    def xml_template(self) -> str:
        return str("""<BreakSpeedMultiplier>{{multiplier}}</BreakSpeedMultiplier>""")

    def __init__(self, multiplier=1.0):
        self.multiplier = multiplier


class AgentStartPlacement(Handler):
    """
    Sets for the agent start location

    Example usage:

    .. code-block:: python

        AgentStartPlacement(x=5, y=70, z=4, yaw=0, pitch=0)
    """

    def to_string(self) -> str:
        return f"agent_start_placement({self.x}, {self.y}, {self.z}, {self.yaw}, {self.pitch})"

    def xml_template(self) -> str:
        return str(
            """<Placement x="{{x}}" y="{{y}}" z="{{z}}" yaw="{{yaw}}" pitch="{{pitch}}"/>"""
        )

    def __init__(self, x, y, z, yaw, pitch=0.0):
        self.x = x
        self.y = y
        self.z = z
        self.yaw = yaw
        self.pitch = pitch


class AgentStartNear(Handler):
    """
    Starts agent near another agent

    Example usage:

    .. code-block:: python

        AgentStartNear("MineRLAgent0", min_distance=2, max_distance=10, max_vert_distance=3)
    """

    def to_string(self) -> str:
        return f"agent_start_near({self.anchor_name}, h {self.min_distance} - {self.max_distance}, v {self.max_vert_distance})"

    def xml_template(self) -> str:
        return str(
            """<NearPlayer>
                    <Name>{{anchor_name}}</Name>
                    <MaxDistance>{{max_distance}}</MaxDistance>
                    <MinDistance>{{min_distance}}</MinDistance>
                    <MaxVertDistance>{{max_vert_distance}}</MaxVertDistance>
                    <LookingAt>true</LookingAt>
               </NearPlayer>"""
        )

    def __init__(
        self,
        anchor_name="MineRLAgent0",
        min_distance=2,
        max_distance=10,
        max_vert_distance=3,
    ):
        self.anchor_name = anchor_name
        self.min_distance = min_distance
        self.max_distance = max_distance
        self.max_vert_distance = max_vert_distance


class StartingHealthAgentStart(Handler):
    """
    Sets the starting health of the agent

    Example usage:

    .. code-block:: python

        StartingHealthAgentStart(max_health=20, health=2.5)

    :code:`max_health` sets the maximum amount of health the agent can have
    :code:`health` sets amount of health the agent starts with (max_health if not specified)
    """

    def to_string(self) -> str:
        return "starting_health_agent_start"

    def xml_template(self) -> str:
        if self.health is None:
            return str("""<StartingHealth maxHealth="{{ max_health }}"/>""")
        else:
            return str(
                """<StartingHealth maxHealth="{{ max_health }}" health="{{ health }}"/>"""
            )

    def __init__(self, max_health: float = 20, health: float = None):
        """ """
        self.health = health
        self.max_health = max_health


class StartingFoodAgentStart(Handler):
    """
    Sets the starting food and/or food saturation of the agent.

    Example usage:

    .. code-block:: python

        StartingFoodAgentStart(food=2.5, food_saturation=1)

    Args:
        :code:`food`: The amount of food the agent starts out with
        :code:`food_saturation`: Determines how fast the hunger level depletes, defaults to 5
    """

    def to_string(self) -> str:
        return "starting_food_agent_start"

    def xml_template(self) -> str:
        return str("""<StartingFood food="{{ food }}"/>""")

    def __init__(self, food: int = 20):
        self.food = food


class RandomizedStartDecorator(Handler):
    def to_string(self) -> str:
        return "randomized_start_decorator"

    def xml_template(self) -> str:
        return str("""<RandomizedStartDecorator/>""")


class LowLevelInputsAgentStart(Handler):
    def to_string(self) -> str:
        return "low_level_inputs"

    def xml_template(self) -> str:
        return "<LowLevelInputs>true</LowLevelInputs>"



# Copyright (c) 2020 All Rights Reserved
# Author: William H. Guss, Brandon Houghton
from minedojo.sim.handler import Handler


class NavigationDecorator(Handler):
    """
    Specifies the navigate goal.
    This class should enable the parameterization of all of the fields in the XML.

    When used to create a Gym environment, they should be passed to create_server_decorators.
    """

    def to_string(self) -> str:
        return "navigation_decorator"

    def xml_template(self) -> str:
        return str(
            """<NavigationDecorator>
                <randomPlacementProperties>
                    <maxRandomizedRadius>{{max_randomized_radius}}</maxRandomizedRadius>
                    <minRandomizedRadius>{{min_randomized_radius}}</minRandomizedRadius>
                    <maxRadius>{{max_radius}}</maxRadius>
                    <minRadius>{{min_radius}}</minRadius>
                    <block>{{block}}</block>
                    <placement>{{placement}}</placement>
                </randomPlacementProperties>
                <minRandomizedDistance>{{min_randomized_distance}}</minRandomizedDistance>
                <maxRandomizedDistance>{{max_randomized_distance}}</maxRandomizedDistance>
                <randomizeCompassLocation>{{randomize_compass_location | string | lower}}</randomizeCompassLocation>
            </NavigationDecorator>
            """
        )

    def __init__(
        self,
        max_randomized_radius: int = 64,
        min_randomized_radius: int = 64,
        min_randomized_distance: int = 0,
        max_randomized_distance: int = 8,
        max_radius: int = 8,
        min_radius: int = 0,
        block: str = "diamond_block",
        placement: str = "fixed_surface",
        randomize_compass_location: bool = False,
    ):
        """Initialize navigation decorator

        :param max_randomized_radius: Maximum value to randomize placement
        :param min_randomized_radius: Minimum value to randomize placement
        :param max_radius: Maximum radius to place in the X axis
        :param min_radius: Minimum radius to place in the X axis
        :param block: Type of block to appear.
        :param placement: 'fixed_surface' or otherwise (see XML schema)
        """
        self.max_randomized_radius = max_randomized_radius
        self.min_randomized_radius = min_randomized_radius
        self.max_radius = max_radius
        self.min_radius = min_radius
        self.block = block
        self.placement = placement
        self.randomize_compass_location = randomize_compass_location
        self.min_randomized_distance = min_randomized_distance
        self.max_randomized_distance = max_randomized_distance



from __future__ import annotations

import glob
import json
import os

from PIL import Image

from .download import download as dl
from .download import get_fn


class WikiDataset:
    """
    Class for MineDojo Wiki Database API.
    We follow PyTorch Dataset format but without actually inheriting from PyTorch dataset to keep the framework general.

    Args:
        download: If ``True`` and there is no existing cache directory, the data will be downloaded automatically.

        download_dir: Directory path where the downloaded data will be saved.
                Default: ``~/.minedojo/``.

        full: If ``True``, the full version of the Wiki database will be downlaoded.
                If ``False``, only a sample version of the Wiki database will be downloaded.
                Default: ``True``.

    Examples:
        >>> from minedojo.data import WikiDataset
        >>> wiki_dataset = WikiDataset()
        >>> print(wiki_dataset[0].keys())
        dict_keys(['metadata', 'tables', 'images', 'sprites', 'texts', 'screenshot'])
    """

    def __init__(
        self,
        *,
        download: bool = True,
        download_dir: None | str = None,
        full: bool = True,
    ):
        if download_dir is None:
            download_dir = os.path.join(os.path.expanduser("~"), ".minedojo")

        if download:
            self.root = dl("wiki", download_dir, full)
        else:
            self.root, _, url = get_fn("wiki", download_dir, full)
            assert os.path.exists(self.root), (
                f"Wiki data folder {self.root} does not exist. "
                "Please set download=True or you can manually "
                f"download and extract it from {url}."
            )

        self.data_paths = glob.glob(f"{self.root}/**/data.json", recursive=True)

    def __len__(self):
        return len(self.data_paths)

    def __getitem__(self, idx):
        data_path = self.data_paths[idx]
        with open(data_path, "r") as f:
            data = json.load(f)

        screenshot_path = data_path.replace("data.json", "screenshot.png")
        with Image.open(screenshot_path) as im:
            im = self._to_rgb(im)
        data["screenshot"] = im

        data_dir = data_path.replace("data.json", "")
        images = []
        for image in data["images"]:
            img_path = f"{data_dir}/{image['path']}"
            with Image.open(img_path) as im:
                im = self._to_rgb(im)
            image["rgb"] = im
            images.append(image)
        data["images"] = images
        return data

    def _to_rgb(self, im: Image.Image):
        if im.mode in ("RGBA", "P"):
            rgba = im.convert("RGBA")
            background = Image.new("RGBA", rgba.size, (255, 255, 255))
            return Image.alpha_composite(background, rgba).convert("RGB")
        else:
            return im.convert("RGB")



"""
These handlers cause the episode to terminate based on certain agent conditions.

When used to create a Gym environment, they should be passed to :code:`create_agent_handlers`
"""
# Copyright (c) 2020 All Rights Reserved
# Author: William H. Guss, Brandon Houghton

from typing import List, Dict, Union

from minedojo.sim.handler import Handler


#  <AgentQuitFromTouchingBlockType>
#     <Block type="diamond_block"/>
#     <Block type="iron_block"/>
# </AgentQuitFromTouchingBlockType>
class AgentQuitFromTouchingBlockType(Handler):
    """
    Terminates episode when agent touches one of the blocks in :code:`blocks`

    Example usage:

    .. code-block:: python

        AgentQuitFromTouchingBlockType([
            "gold_block", "oak_log"
        ])
    """

    def to_string(self) -> str:
        return "agent_quit_from_touching_block_type"

    def xml_template(self) -> str:
        return str(
            """<AgentQuitFromTouchingBlockType>
                    {% for block in blocks %}
                    <Block type="{{ block }}"/>
                    {% endfor %}
                </AgentQuitFromTouchingBlockType>"""
        )

    def __init__(self, blocks: List[str]):
        """Creates a reward which will cause the player to quit when they touch a block."""
        self.blocks = blocks


#  <AgentQuitFromDeath/>
class AgentQuitFromDeath(Handler):
    def to_string(self) -> str:
        return "agent_quit_from_death"

    def xml_template(self) -> str:
        return str("""<AgentQuitFromDeath/>""")


# <AgentQuitFromCraftingItem>
#     <Item type="iron_pickaxe"/>
#     <Item type="wooden_axe"/>
#     <Item type="chest"/>
# </AgentQuitFromCraftingItem>
class AgentQuitFromCraftingItem(Handler):
    """
    Terminates episode when agent crafts one of the items in :code:`items`

    Example usage:

    .. code-block:: python

        AgentQuitFromCraftingItem([
            dict(type="iron_axe", amount=1), dict(type="diamond_block", amount=5)
        ])
    """

    def to_string(self) -> str:
        return "agent_quit_from_crafting_item"

    def xml_template(self) -> str:
        return str(
            """<AgentQuitFromCraftingItem>
                    {% for item in items %}
                    <Item type="{{ item.type}}" amount="{{ item.amount }}"/>
                    {% endfor %}
                </AgentQuitFromCraftingItem>"""
        )

    def __init__(self, items: List[Dict[str, Union[str, int]]]):
        """Creates a reward which will cause the player to quit when they have finished crafting something."""
        self.items = items

        for item in self.items:
            assert "type" in item, "{} does contain `type`".format(item)
            assert "amount" in item, "{} does not contain `amount`".format(item)


#  <AgentQuitFromPossessingItem>
#     <Item type="log" amount="64"/>
# </AgentQuitFromPossessingItem>
class AgentQuitFromPossessingItem(Handler):
    """
    Terminates episode when agent obtains one of the items in :code:`items`

    Example usage:

    .. code-block:: python

        AgentQuitFromPossessingItem([
            dict(type="golden_apple", amount=3), dict(type="diamond", amount=1)
        ])
    """

    def to_string(self) -> str:
        return "agent_quit_from_possessing_item"

    def xml_template(self) -> str:
        return str(
            """<AgentQuitFromPossessingItem>
                   {% for item in items %}
                   <Item type="{{ item.type }}" amount="{{ item.amount }}"/>
                   {% endfor %}
               </AgentQuitFromPossessingItem>"""
        )

    def __init__(self, items: List[Dict[str, Union[str, int]]]):
        assert isinstance(items, list)
        self.items = items
        # Assert that all the items have the correct fields for the XML.
        for item in self.items:
            assert "type" in item, "{} does contain `type`".format(item)
            assert "amount" in item, "{} does not contain `amount`".format(item)



from .sim import MineDojoSim
from .inventory import InventoryItem
from .mc_meta.mc import (
    ALL_ITEMS,
    ALL_ITEM_VARIANTS,
    ALL_CRAFTING_TABLE_ITEMS,
    ALL_CRAFTING_TABLE_ONLY_ITEMS,
    ALL_PERSONAL_CRAFTING_ITEMS,
    ALL_CRAFT_SMELT_ITEMS,
    ALL_SMELTING_ITEMS,
    ALL_STATS,
    BIOMES_MAP,
)



# Copyright (c) 2020 All Rights Reserved
# Author: William H. Guss, Brandon Houghton
import minedojo.sim.spaces as spaces
from minedojo.sim.handlers.agent.action import Action


class KeybasedCommandAction(Action):
    """
    A command action which is generated from human keypresses in anvil.
    Examples of such actions are movement actions, etc.

    This is not to be confused with keyboard acitons, wehreby both anvil and malmo
    simulate and act on direct key codes.

    Combinations of KeybasedCommandActions yield actions like:

    .. code-block:: json

        {
            “move” : 1,
            “jump”: 1
        }
    where move and jump are the commands, which correspond to keys like 'W', 'SPACE', etc.

    This is as opposed to keyboard actions (see the following class definition in keyboard.py)
    which yield actions like:

    .. code-block:: json

        {
            "keyboard" : {
                "W" : 1,
                "A": 1,
                "S": 0,
                "E": 1,
                ...
            }
        }
    More information can be found in the unification document (internal).
    """

    def to_string(self):
        return self.command

    def xml_template(self) -> str:
        """Notice how all of the instances of keybased command actions,
        of which there will be typically many in an environment spec,
        correspond to exactly the same XML stub.

        This is discussed at length in the unification proposal
        and is a chief example of where manifest consolidation is needed.
        """
        return str("<HumanLevelCommands/>")

    def __init__(self, command, *keys):
        if len(keys) == 2:
            # Like move or strafe. Example: -1 for left, 1 for right
            super().__init__(command, spaces.DiscreteRange(-1, 2))
        else:
            # Its a n-key action with discrete items.
            # Eg hotbar actions
            super().__init__(command, spaces.Discrete(len(keys) + 1))
        self.keys = keys

    def from_universal(self, x):
        # actions_mapped is just the raw key codes.
        actions_mapped = list(x["custom_action"]["actions"].keys())
        offset = self.space.begin if isinstance(self.space, spaces.DiscreteRange) else 0
        default = 0

        for i, key in enumerate(self.keys):
            if key in actions_mapped:
                if isinstance(self.space, spaces.DiscreteRange):
                    return i * 2 + offset
                else:
                    return i + 1 + offset

        return default


# TODO: This will be useful for when full keyboard actions are introduced.
# class KeyboardAction(TranslationHandler):
#     """Enables a set of keyboard actions.
#     This handler correspomds direcrtly with the HumanLevelCommands handler
#     in MissionHandlers.xsd.

#     """
#     def xml_template(self) -> str:
#         return str("""
#             <HumanLevelCommands>
#                 <ModifierList type="allow-list">
#                     {% for command in commands %}
#                         <command>{{ command }}</command>
#                     {% endfor %}
#                 </ModifierList>
#             </HumanLevelCommands>
#             """)


#     def to_string(self) -> str:
#         return "keyboard"

#     def __init__(self, keymap : typing.Dict[str, str]):
#         """Initializes the keyboard action object with a keymap
#         Anvil rendered keypress ID's and Malmo command actions corresponding to the human level commands object.

#         Args:
#             keymap (Dict[str, str]): A keymap between anvil keys & Malmo commands (see KEYMAP)
#         """
#         self.keymap = keymap

#         super().__init__(spaces.Dict(
#             {
#                 v: spaces.Discrete(1) for v in keymap.values()
#             }
#         ))

#     def xml(self) -> str:
#         return self.TEMPLATE.render(commands=self.keymap.values())

#     def to_hero(self, x : typing.Dict[str, np.ndarray]) -> str:
#         """ Joins all of the commands in X to a string by new lines.
#         First joins the keys and values in X with a space.

#         Args:
#             x (typing.Dict[str, Any]): A KeyboardAction.
#         """
#         return   '\n'.join(['%s %s' % (k, v) for k, v in x.items()])

#     def from_universal(self, x: typing.Dict[str, Any]) -> typing.Dict[str, np.ndarray]:
#         """Finds all of the keys in the universal json corresponding to keypresses
#         and if they are present produces an observation dictionary with keymap keys
#         having value 1 and otherwise 0.
#         """
#         actions_mapped = list(x['custom_action']['actions'].keys())
#         # actions_mapped is just the raw key codes.

#         out = self.space.no_op()
#         out.update({
#             out[self.keymap[a]] : np.array(1, dtype=np.int) for a in actions_mapped
#         })


#         return out

#     def __or__(self, other):
#         """Combines the keymaps from both dealios.
#         """
#         super().__or__(other)
#         new_keymap = {}
#         new_keymap.update(self.keymap)
#         new_keymap.update(other.keymap)
#         return  KeyboardAction(keymap=new_keymap)



# Copyright (c) 2020 All Rights Reserved
# Author: William H. Guss, Brandon Houghton

from .mc_base_stats import *
from .location_stats import *
from .compass import *
from .lifestats import *
from .pov import *
from .damage_source import *
from .voxel_lidar import VoxelObservation, RichLidarObservation
from .true_flat_inventory import TrueFlatInventoryObservation, EquipmentObservation
from .achievements import ObsFromAchievements
from .nearby_tools import NearbyToolsObservation



from typing import Dict, Any

from minedojo.sim import spaces
from minedojo.sim.mc_meta.mc import ALL_ACHIEVEMENTS
from minedojo.sim.handlers.translation import TranslationHandler


class ObsFromAchievements(TranslationHandler):
    def to_string(self) -> str:
        return "achievements"

    def xml_template(self) -> str:
        return "<ObservationFromAchievements/>"

    def __init__(self):
        super().__init__(
            space=spaces.Dict(
                {achievement: spaces.Discrete(n=2) for achievement in ALL_ACHIEVEMENTS}
            )
        )

    def from_hero(self, x: Dict[str, Any]):
        return {k: int(v) for k, v in x[self.to_string()].items()}



import struct


def send_message(sock, data):
    length = len(data)
    sock.sendall(struct.pack("!I", length))
    sock.sendall(data)


def recv_message(sock):
    lengthbuf = _recv_all(sock, 4)
    if not lengthbuf:
        return None
    (length,) = struct.unpack("!I", lengthbuf)
    return _recv_all(sock, length)


def _recv_all(sock, count):
    """
    receive all
    """
    buf = b""
    while count:
        newbuf = sock.recv(count)
        if not newbuf:
            return None
        buf += newbuf
        count -= len(newbuf)
    return buf



materials = [
    "diamond",
    "gold",
    "iron",
    "stone",
    "glass",
    "sand",
]
architectures_farm = [
    "house",
    "castle",
    "temple",
    "pyramid",
    "tower",
]

architectures = [
    "ice house",
    "sand castle",
    "glass temple",
    "dirt pyramid",
    "wooden tower",
    "stone cathedral",
    "iron hospital",
    "gold school",
    "diamond skyscraper",
]
statues = [
    "diamond car",
    "glod train",
    "iron plane",
    "stone apple",
    "wooden dog",
    "glass cat",
    "ice man",
    "sand woman",
]

famous_sight = [
    "Grand Canyon",
    "Roman Colosseum",
    "Eiffel Tower",
    "Taj Mahal",
    "Notre Dame",
    "Great Wall",
    "Statue of Liberty",
    "Cologne Cathedral",
    "Sydney Opera House",
    "Acropolis",
    "Golden Gate Bridge",
    "Leaning Tower of Pisa",
]

natural_scenes = [
    "waterfall",
    "lava fall",
    "tree",
    "cave",
    "mountain",
    "lake",
    "lava lake",
]

view_point_2d = ["top-down", "vertical", "panorama", "half-panorama"]

passive_mobs = ["pig", "sheep", "cow", "chicken"]
hostile_mobs = ["zombie", "creeper", "skeleton", "spider"]

automatic_farms = [
    "wheat",
    "gold",
    "cow",
    "sheep",
    "sugar cane",
    "hostile mob",
    "wool",
]

explore_targets = [
    "plains",
    "forest",
    "ocean",
    "river",
    "beach",
    "taiga",
    "mountains",
    "ice spikes",
    "jungle",
    "desert",
    "swampland",
    "end city",
    "ocean monument",
    "desert temple",
    "woodland mansion",
    "cave",
    "mineshaft",
    "swamp hut",
    "jungle temple",
    "mushroom island",
    "nether",
    "village",
    "pyramid",
    "cliff",
    "waterfall",
    "lavafall",
    "lake",
    "lava lake",
    "underground lake",
]

all_task = {
    "explore": dict(
        template="Explore the world to find {explore_targets}.",
        parameters=dict(
            explore_targets=explore_targets,
        ),
    ),
    "architecture": dict(
        template="Build a {architectures}.",
        parameters=dict(
            architectures=architectures,
        ),
    ),
    "farm": dict(
        template="Build a {passive_mobs} farm.",
        parameters=dict(
            passive_mobs=passive_mobs,
        ),
    ),
    "farm_architectures": dict(
        template="Build a {passive_mobs} farm in {architectures_farm}.",
        parameters=dict(
            passive_mobs=passive_mobs,
            architectures_farm=architectures_farm,
        ),
    ),
    "natural_scenes": dict(
        template="Build an artificial {natural_scenes}.",
        parameters=dict(natural_scenes=natural_scenes),
    ),
    "statue": dict(
        template="Build a {statues} statue.",
        parameters=dict(statues=statues),
    ),
    "famous_sight": dict(
        template="Build the {famous_sight}.",
        parameters=dict(
            famous_sight=famous_sight,
        ),
    ),
    "automatic_farm": dict(
        template="Build an automatic {resource} farm.",
        parameters=dict(
            resource=automatic_farms,
        ),
    ),
    "haunted_house": dict(
        template="Build a Haunted House with {hostile_mobs} inside.",
        parameters=dict(
            hostile_mobs=hostile_mobs,
        ),
    ),
    "swimming_pool": dict(
        template="Build a swimming pool.",
        parameters=dict(),
    ),
    "maze": dict(
        template="Build a maze using {materials}.",
        parameters=dict(materials=materials),
    ),
    "archery_range": dict(
        template="Build an archery range.",
        parameters=dict(),
    ),
    "bridge": dict(
        template="Build a bridge over a {fluid}.",
        parameters=dict(
            fluid=["river", "lake", "lava"],
        ),
    ),
    "light_tree": dict(
        template="Build a tree that can be lighted up.",
        parameters=dict(),
    ),
    "trap": dict(
        template="Trap a {hostile_mobs} in {containers}.",
        parameters=dict(
            hostile_mobs=hostile_mobs,
            containers=architectures_farm + ["wooden cage", "stone cage"],
        ),
    ),
    "gather": dict(
        template="Gather nearby {passive_mobs} in a {architectures_farm}.",
        parameters=dict(
            passive_mobs=passive_mobs,
            architectures_farm=architectures_farm,
        ),
    ),
    "travel_long_distance": dict(
        template="Travel over {distance} blocks by {approach}.",
        parameters=dict(
            distance=[1000, 2000, 3000],
            approach=[
                "ice boat",
                "nether portals",
                "mine cart",
                "riding horse",
                "riding pig",
            ],
        ),
    ),
    "fill_map": dict(
        template="Fill a map of the world.",
        parameters=dict(),
    ),
    "reach_upperbound": dict(
        template="Reach the upperbound of the world.",
        parameters=dict(),
    ),
    "race": dict(
        template="Race by {approach}.",
        parameters=dict(approach=["ice boat", "boat", "riding horse", "riding pig"]),
    ),
    "bring_in_nether": dict(
        template="Bring a {passive_mobs} into nether.",
        parameters=dict(passive_mobs=passive_mobs),
    ),
    "sail": dict(
        template="Sail on boat with a {passive_mobs}.",
        parameters=dict(passive_mobs=passive_mobs),
    ),
    "tame": dict(
        template="Tame a {tamable_animal}.",
        parameters=dict(tamable_animal=["horse", "wolf"]),
    ),
    "light_up": dict(
        template="light up a {architectures}.",
        parameters=dict(architectures=architectures_farm + ["cave"]),
    ),
    "clog": dict(
        template="Clog a {fluid_source}.",
        parameters=dict(fluid_source=["waterfall", "lava fall", "river"]),
    ),
    "climb": dict(
        template="Climb a waterfall.",
        parameters=dict(),
    ),
    "hook": dict(
        template="Hook a {mobs} using fishing rod.",
        parameters=dict(mobs=passive_mobs + hostile_mobs),
    ),
}
